#ifndef RAMTRON_H
#define RAMTRON_H

#include <system.h>
#include "uart.h"
#include "util.h"
#include <stdio.h>

#define FM3164_MEMORY_SIZE 0xffff
#define FM3164_CURRENT_LOCATION 0xffff

// Uncomment to enable low-level I2C read/write debug output.
//#define FM3164_DEBUG 1

// Macro for simplifying waiting for an idle condition on the i2c interface.
#define I2C_IDLE_WAIT() while ((sspcon2 & 0x1F) | (sspstat.R_W))

typedef enum i2c_status
{
	I2C_NO_ERROR = 0,
	I2C_WRITE_COLLISION,
	I2C_NO_ACK
} i2c_status_t;

typedef enum fm3164_registers
{
	FM3164_REGISTER_SERIAL7 = 0x18,
	FM3164_REGISTER_SERIAL6 = 0x17,
	FM3164_REGISTER_SERIAL5 = 0x16,
	FM3164_REGISTER_SERIAL4 = 0x15,
	FM3164_REGISTER_SERIAL3 = 0x14,
	FM3164_REGISTER_SERIAL2 = 0x13,
	FM3164_REGISTER_SERIAL1 = 0x12,
	FM3164_REGISTER_SERIAL0 = 0x11,
	FM3164_REGISTER_COUNTER2H = 0x10,
	FM3164_REGISTER_COUNTER2L = 0x0f,
	FM3164_REGISTER_COUNTER1H = 0x0e,
	FM3164_REGISTER_COUNTER1L = 0x0d,
	FM3164_REGISTER_COUNTER_CONTROL = 0x0c,
	FM3164_REGISTER_COMPANION_CONTROL= 0x0b,
	FM3164_REGISTER_WATCHDOG_CONTROL = 0x0a,
	FM3164_REGISTER_WATCHDOG = 0x09,
	FM3164_REGISTER_YEAR = 0x08,
	FM3164_REGISTER_MONTH = 0x07,
	FM3164_REGISTER_DAY = 0x06,
	FM3164_REGISTER_DAYOFWEEK = 0x05,
	FM3164_REGISTER_HOURS = 0x04,
	FM3164_REGISTER_MINUTES = 0x03,
	FM3164_REGISTER_SECONDS = 0x02,
	FM3164_REGISTER_CONTROL = 0x01,
	FM3164_REGISTER_FLAGS = 0x00
} fm3164_registers_t;

typedef struct fm3164_date
{
	unsigned char year;
	unsigned char month;
	unsigned char day;
	unsigned char hours;
	unsigned char minutes;
	unsigned char seconds;
} fm3164_date_t;

void i2c_init (void);
void fm3164_init (void);

i2c_status_t i2c_write (unsigned char data);

i2c_status_t fm3164_write_register (
	unsigned char address, unsigned char data);

i2c_status_t fm3164_read_register (
	unsigned char address, unsigned char *data);

i2c_status_t fm3164_read_memory (unsigned int address,
	unsigned char *data, unsigned char length);

i2c_status_t fm3164_read_memory_byte (unsigned int address,
	unsigned char *dst);

i2c_status_t fm3164_write_memory (unsigned int address,
	unsigned char *data, unsigned char length);

i2c_status_t fm3164_write_byte_memory (unsigned int address,
	unsigned char data);

void i2c_get_bytes (unsigned char *data, unsigned char length);

void fm3164_get_date (fm3164_date_t *date);
void fm3164_set_date (fm3164_date_t *date);

void fm3164_date_to_compact (const fm3164_date_t *src, unsigned int *dst);

#endif