#include <system.h>

#pragma CLOCK_FREQ 40000000

#define DISPLAY_D4_DDR trisa.RA0
#define DISPLAY_D5_DDR trisa.RA1
#define DISPLAY_D6_DDR trisa.RA2
#define DISPLAY_D7_DDR trisa.RA3
#define DISPLAY_RS_DDR trisa.RA5
#define DISPLAY_RW_DDR trisc.RC0
#define DISPLAY_E_DDR trisc.RC1

#define DISPLAY_D4 lata.RA0
#define DISPLAY_D5 lata.RA1
#define DISPLAY_D6 lata.RA2
#define DISPLAY_D7 lata.RA3
#define DISPLAY_RS lata.RA5
#define DISPLAY_RW latc.RC0
#define DISPLAY_E latc.RC1

#define DISPLAY_DATA lata

typedef enum display_mode
{
	DISPLAY_MODE_1_LINE_SMALL_FONT,
	DISPLAY_MODE_1_LINE_LARGE_FONT,
	DISPLAY_MODE_2_LINE_SMALL_FONT
} display_mode_t;


void display_4bit_init (display_mode_t mode);
void init2 (void);
void init3 (void);
void display_4bit_set_position (unsigned char x, unsigned char y);
void display_4bit_write_character (char character);
void display_4bit_write_string (const char *string);

// Internal functions (only visible in this file).
static void display_4bit_write_byte (unsigned char command);
static void display_4bit_write_nibble (unsigned char command);

void main (void)
{
	unsigned char i;

	delay_ms(250);

	// Configure I/O pins.
	DISPLAY_D4_DDR = 0;
	DISPLAY_D5_DDR = 0;
	DISPLAY_D6_DDR = 0;
	DISPLAY_D7_DDR = 0;

	DISPLAY_E_DDR = 0;
	DISPLAY_RS_DDR = 0;
	DISPLAY_RW_DDR = 0;
	
	DISPLAY_RS = 0;
	DISPLAY_RW = 0;
	DISPLAY_E = 0;

	delay_us(2);

	display_4bit_write_nibble(0x02);
	delay_us(60);
	display_4bit_write_nibble(0x02);
	display_4bit_write_nibble(0x08);
	delay_us(60);
	display_4bit_write_nibble(0x00);
	display_4bit_write_nibble(0x0c);
	delay_us(60);

	//display_4bit_init(DISPLAY_MODE_2_LINE_SMALL_FONT);
	//init2();
	//init3();

	

/*
	display_4bit_set_position(5, 1);
	display_4bit_write_string("Some");
	display_4bit_set_position(7, 0);
	display_4bit_write_string("things");
*/

	for (i = 0; i < 0x80; ++i) {
		display_4bit_write_character('x');
	}
	

	while(1);	
}

void display_4bit_init (display_mode_t mode)
{
	display_4bit_write_nibble(0x03);

	delay_ms(5);

	display_4bit_write_nibble(0x03);

	delay_us(120);

	display_4bit_write_nibble(0x03);

	delay_ms(5);

	delay_us(40);

	display_4bit_write_byte(0x22);

	delay_us(40);

	display_4bit_write_byte(0x0c);

	delay_ms(5);

	display_4bit_write_byte(0x14);

	delay_us(40);
}

void init2 (void)
{
	unsigned char temp;

	// First set us to 8-bit mode.
	display_4bit_write_nibble(0x03);

	delay_ms(40);

	display_4bit_write_nibble(0x02);

	delay_ms(40);

	// Function set, 4-bit, 2-line, 5x7 font.
	display_4bit_write_byte(0x28);
	delay_us(40);

	// Cursor display config, shift cursor left.
	display_4bit_write_byte(0x14);
	delay_us(40);

	// Display on/off, enable display but no cursor or blink.
	display_4bit_write_byte(0x0c);
	delay_us(40);

	// Entry mode, right direction with no shift.
	display_4bit_write_byte(0x06);
	delay_us(40);
}

void init3 (void)
{
	display_4bit_write_nibble(0x02);
	display_4bit_write_nibble(0x02);
	display_4bit_write_nibble(0x0c);

	delay_us(40);

	display_4bit_write_byte(0x0c);

	delay_us(40);

	display_4bit_write_byte(0x01);

	delay_ms(2);

	display_4bit_write_byte(0x02);

	delay_us(40);
}

void display_4bit_set_position (unsigned char x, unsigned char y)
{
	if (y == 1) { x += 0x40; }

	DISPLAY_RS = 0;
	DISPLAY_RW = 0;

	display_4bit_write_byte(x | 0x80);
}

void display_4bit_write_character (char character)
{
	// Configure signals for outgoing character data.
	DISPLAY_RS = 1;
	DISPLAY_RW = 0;

	display_4bit_write_byte(character);

	delay_us(40);
}

void display_4bit_write_string (const char *string)
{
	// This code can be optimized further but this is
	// generally more compatible with crappy compilers.
	char *p;

	DISPLAY_RS = 1;
	DISPLAY_RW = 0;

	for (p = string; *p != '\0'; ++p) {
		display_4bit_write_byte(*p);
	}

}

void display_4bit_write_byte (unsigned char command)
{
	display_4bit_write_nibble((command >> 4) & 0x0f);
	display_4bit_write_nibble(command & 0x0f);
}

void display_4bit_write_nibble (unsigned char command)
{
	unsigned char temp;

	// Write the LSB nibble.
	temp = DISPLAY_DATA;
	temp &= command | 0xf0;
	temp |= command & 0x0f;
	DISPLAY_DATA = temp;
	delay_us(2);

	// Flash 'enable' high to transfer data.
	DISPLAY_E = 1;
	delay_us(5);
	DISPLAY_E = 0;
	delay_us(5);
}