#include <p18f2515.h>
#include <delays.h>

// Pin mappings / register mappings
#define DISPLAY_D4 LATAbits.LATA0
#define DISPLAY_D5 LATAbits.LATA1
#define DISPLAY_D6 LATAbits.LATA2
#define DISPLAY_D7 LATAbits.LATA3

#define DISPLAY_E  LATCbits.LATC1
#define DISPLAY_RS LATAbits.LATA5
#define DISPLAY_RW LATCbits.LATC0

void init_elektronikforumet (void);
void init_sitronix (void);

void send_command (unsigned char command);
void delay_us (unsigned char microseconds);
void delay_ms (unsigned char milliseconds);
void display_write_character (unsigned char character);

void main (void)
{
	TRISA = 0x00;
	TRISC = 0x00;
	LATA = 0x00;
	LATC = 0x00;
	ADCON1 |= 0x0f;

	init_elektronikforumet();
	//init_sitronix();

	display_write_character('c');

	while (1);	
}

void init_elektronikforumet (void)
{
	delay_ms(250);

	DISPLAY_E = 0;
	DISPLAY_RS = 0;
	DISPLAY_RW = 0;

	send_command(0x03);

	delay_ms(6);

	send_command(0x03);

	delay_us(120);

	send_command(0x03);
	send_command(0x02);
	delay_us(50);

	send_command(0x02);
	send_command(0x08);
	delay_us(50);

	send_command(0x00);
	send_command(0x80);

	send_command(0x00);
	send_command(0x01);
	delay_us(50);

	send_command(0x00);
	send_command(0x06);
	delay_us(50);
}

void init_sitronix (void)
{
	delay_ms(250);

	DISPLAY_E = 0;
	DISPLAY_RS = 0;
	DISPLAY_RW = 0;

	send_command(0x03);

	delay_us(50);

	send_command(0x02);
	send_command(0x02);

	delay_us(50);

	send_command(0x02);
	send_command(0x02);

	delay_us(50);

	send_command(0x00);
	send_command(0x0c);

	delay_us(50);

	send_command(0x00);
	send_command(0x01);

	delay_us(50);

	send_command(0x00);
	send_command(0x06);

	delay_us(50);
}

void send_command (unsigned char command)
{
	unsigned char temp;

	temp = LATA & 0xf0;
	temp |= command & 0x0f;
	LATA = temp;
	delay_us(2);

	DISPLAY_E = 1;
	delay_us(6);
	DISPLAY_E = 0;
	delay_us(6);
}

void delay_us (unsigned char microseconds)
{
	Delay10TCYx(microseconds);
}

void delay_ms (unsigned char milliseconds)
{
	Delay10KTCYx(milliseconds);
}

void display_write_character (unsigned char character)
{
	unsigned char temp;

	DISPLAY_RS = 1;
	DISPLAY_RW = 1;
	delay_us(2);

	temp = LATA;
	temp &= (character >> 4) | 0xf0;
	temp |= (character >> 4) & 0x0f;
	LATA = temp;

	DISPLAY_E = 1;
	delay_us(6);
	DISPLAY_E = 0;
	delay_us(6);

	temp = LATA;
	temp &= character | 0xf0;
	temp |= character & 0x0f;
	LATA = temp;

	DISPLAY_E = 1;
	delay_us(6);
	DISPLAY_E = 0;
	delay_us(6);

	delay_us(50);
}