#include "display_4bit.h"

// Internal functions (only visible in this file).
static void display_4bit_write_byte (unsigned char command);
static void display_4bit_write_nibble (unsigned char command);

void display_4bit_init (void)
{
	// Issue 'function set' in 8-bit mode, changing interface
	// to 4-bits from now on.
	display_4bit_write_nibble(0x02);
	delay_us(60);

	// Issue 'function set' in 4-bit mode, changing the number
	// of lines on the display.
	display_4bit_write_nibble(0x02);
	display_4bit_write_nibble(0x08);
	delay_us(60);

	// Issue 'display on' to turn the display back on after change.
	display_4bit_write_nibble(0x00);
	display_4bit_write_nibble(0x0c);
	delay_us(60);

	// Clear the display.
	display_4bit_write_nibble(0x00);
	display_4bit_write_nibble(0x01);
	delay_ms(10);

	// Entry mode set.
	//display_4bit_write_nibble(0x00);
	//display_4bit_write_nibble(0x06);
}

void display_4bit_set_position (unsigned char x, unsigned char y)
{
	if (y == 1) { x += 0x40; }

	DISPLAY_RS = 0;
	DISPLAY_RW = 0;

	display_4bit_write_byte(x | 0x80);
	//display_4bit_write_nibble((x | 0x80) >> 4);
	//display_4bit_write_nibble(x);
	delay_us(60);
}

void display_4bit_write_character (char character)
{
	// Configure signals for outgoing character data.
	DISPLAY_RS = 1;
	DISPLAY_RW = 0;

	//display_4bit_write_byte(character);
	display_4bit_write_nibble(character >> 4);
	display_4bit_write_nibble(character);

	delay_us(60);
}

void display_4bit_write_string (const char *string)
{
	// This code can be optimized further but this is
	// generally more compatible with crappy compilers.
	char *p;

	DISPLAY_RS = 1;
	DISPLAY_RW = 0;

	for (p = string; *p != '\0'; ++p) {
		//display_4bit_write_nibble(*p >> 4);
		//display_4bit_write_nibble(*p);
		display_4bit_write_byte(*p);
		delay_us(60);
	}

}

void display_4bit_clear (void)
{
	display_4bit_write_byte(0x01);
	delay_ms(10);
}

void display_4bit_write_byte (unsigned char command)
{
	display_4bit_write_nibble(command >> 4);
	display_4bit_write_nibble(command);

}

void display_4bit_write_nibble (unsigned char command)
{
	unsigned char temp;

	// Write the LSB nibble.
	temp = DISPLAY_DATA;
	temp &= command | 0xf0;
	temp |= command & 0x0f;
	DISPLAY_DATA = temp;

	//DISPLAY_DATA &= 0xf0;
	//DISPLAY_DATA += (0x0f & command);
	
	delay_us(50);

	// Flash 'enable' high to transfer data.
	DISPLAY_E = 1;
	delay_us(100);
	DISPLAY_E = 0;
	delay_us(100);
}