#include "timeit_menu2.h"

//system_states_t system_state;

// Main counter, used for all timing operations.
// Updated async. by the timer interrupt. We need a mutex when accessing this variable.
// Such a mutex is currently NOT implemented.
volatile unsigned long counter_ms = 0;

menu_states_t menu_state = MENU_STATE_SPRINT;

volatile record_t global_record; // The default record where EEPROM data is stuffed when read.
volatile unsigned int current_record = 0x0002; // Current read position for EEPROM operations.
volatile unsigned int current_record_size;

// Common registers used to store information between sessions.
volatile unsigned char register0;
volatile unsigned char register1;
volatile unsigned char register2;
volatile unsigned char register3;
volatile unsigned int register4;
volatile unsigned int register5;
volatile unsigned long results[MAX_SENSORS]; // Array storing the measured times.
volatile unsigned int distances[MAX_SENSORS];

// This array is used when results are being viewed. It caches the addresses of each record.
volatile unsigned int result_addresses[NUM_CACHED_RESULTS];

void main (void)
{
	char buffer[14];
	keys_t key = KEY_NONE;
	unsigned char i;
	unsigned int value;

	delay_s(1);

	// Configure I/O pins for the display.
	DISPLAY_D4_DDR = 0;
	DISPLAY_D5_DDR = 0;
	DISPLAY_D6_DDR = 0;
	DISPLAY_D7_DDR = 0;

	DISPLAY_E_DDR = 0;
	DISPLAY_RS_DDR = 0;
	DISPLAY_RW_DDR = 0;
	
	DISPLAY_RS = 0;
	DISPLAY_RW = 0;
	DISPLAY_E = 0;

	trisa.RA4 = 1; // PortA[4] as an input for mat detection.
	latc.RC2 = 0;
	trisc.RC2 = 0; // Piezo buzzer, output.
	trisb.RB1 = 1; // Pad input.
	trisb.RB0 = 1; // Radio input

/*
	for ( ; ; ) {
		latc.RC2 = 1;
		delay_ms(20);
		latc.RC2 = 0;
		delay_ms(20);
	}
*/
	// Enable i2c interface module.
	i2c_init();

	// Config timer0, 8 bit, prescaler 16, match 125.
	// This will generate an interrupt each millisecond.
	t2con = 0x06; // Enable timer, 16 prescaler
	pr2 = 125;
	intcon |= 0xc0;
	pie1 |= 0x02;
	
	delay_us(2);

	display_4bit_init();

	menu_draw();

	uart_enable();

	#ifdef DEBUG
	// Debug initalization code, clears EEPROM memory.
	init_eeprom();

	uart_send_string("Debug enabled, main startup.\r\n");
	#endif

	fm3164_init();
	
	#ifdef DEBUG_WRITE_RECORDS
	global_record.player_id = 0;
	//global_record.test_id = 0xba12;
	fm3164_get_date(&global_record.date);
	global_record.type = TEST_TYPE_SPRINT_SIGNAL;
	global_record.num_times = 2;

	results[0] = 3600000;
	results[1] = 0xb3b2b1b0;

	distances[0] = 0xbeef;
	distances[1] = 0xdead;

	delay_s(1);

	record_store();

	global_record.player_id = 0;
	//global_record.test_id = 0xba13;
	fm3164_get_date(&global_record.date);
	global_record.type = TEST_TYPE_SPRINT_SIGNAL;
	global_record.num_times = 3;

	results[0] = 3600;
	results[1] = 3600000;
	results[2] = 0xb3b2b1b1;

	distances[0] = 0xbef0;
	distances[1] = 0xdeae;
	distances[2] = 0xdeae;

	delay_s(2);

	record_store();

	global_record.player_id = 0;
	global_record.test_id = 0xba13;
	fm3164_get_date(&global_record.date);
	global_record.type = TEST_TYPE_JUMP_SINGLE;
	global_record.num_times = 1;

	results[0] = 960;
	distances[0] = 0x0000;

	delay_s(2);

	record_store();

	global_record.player_id = 0;
	global_record.test_id = 0xba13;
	fm3164_get_date(&global_record.date);
	global_record.type = TEST_TYPE_JUMP_SINGLE;
	global_record.num_times = 1;

	results[0] = 550;
	distances[0] = 0x0000;

	delay_s(2);

	record_store();

	global_record.player_id = 1;
	global_record.test_id = 0xba14;
	fm3164_get_date(&global_record.date);
	global_record.type = TEST_TYPE_SPRINT_SIGNAL;
	global_record.num_times = 2;

	results[0] = 0xcafebabe;
	results[1] = 0xbaadf00d;

	distances[0] = 0xfeed;
	distances[1] = 0xface;

	record_store();
	#endif

	// The main loop.
	while(1) {
		if ((key = keypad_scan()) != KEY_NONE) {
			menu_process(key);
			menu_draw();
			wait_for_keypad_release();
		}
		delay_ms(KEY_PRESS_DELAY);
	}

	// Execution should never reach this point.
}

/*
Function: test_jump_single

Description: Performs a single jump.

Return value: Returns nonzero if testing should continue, zero if not.

*/
unsigned char test_jump_single (void)
{
	unsigned long temp;
	keys_t key;
	unsigned char buffer[8];
	unsigned long timer;

	#ifdef DEBUG_JUMP_SINGLE
	unsigned char buffer2[16];
	#endif

	#ifdef DEBUG_JUMP_SINGLE
	uart_send_string("jump_single\r\n");
	#endif

	global_record.type = TEST_TYPE_JUMP_SINGLE;
	global_record.player_id = register3;

	wait_for_keypad_release();

	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("JUMP!");

	// Store the current time.
	fm3164_get_date(&global_record.date);

	// Wait for the user to land.
	while (portb.RB1) {
		key = keypad_scan();
		if (key == KEY_CANCEL) { return 1; }
		else if (key == KEY_PREV) { return 2; }
	}
	counter_ms = 0;

	delay_ms(100);

	#ifdef DEBUG_JUMP_SINGLE
	uart_send_string("Wait f. jump\r\n");
	#endif

	// Wait for the user to jump.
	while (!portb.RB1) {
		key = keypad_scan();
		if (key == KEY_CANCEL) { return 1; }
		else if (key == KEY_PREV) { return 2; }	
	}
	results[0] = counter_ms;
	counter_ms = 0;

	#ifdef DEBUG_JUMP_SINGLE
	uart_send_string("Wait f. land\r\n");
	#endif

	display_4bit_clear();
	display_4bit_set_position(0, 0);

	// Wait for mat to get depressed again.
	while (portb.RB1) {
		key = keypad_scan();
		if (key == KEY_CANCEL) { return 1; }
		else if (key == KEY_PREV) { return 2; }

		// Display time on col 0, row 1.	
		realtime_update(counter_ms, 0, 1);
	}
	#ifdef DEBUG_JUMP_SINGLE
	uart_send_string("landed!\r\n");
	#endif

	timer = counter_ms;

	#ifdef DEBUG_JUMP_SINGLE
	uart_send_string("airtime: ");
	sprintf32(buffer2, "%u", timer);
	uart_send_string(buffer2);
	uart_send_string("\r\n");

	uart_send_string("grndtime: ");
	sprintf32(buffer2, "%u", results[0]);
	uart_send_string(buffer2);
	uart_send_string("\r\n");
	#endif

	global_record.num_times = 1;
	
	// result = 9.82 (gravity) * (air time (in seconds) / 2) ^ 2
	// TODO: check if this is correct!
	temp = (unsigned long) (timer / 2);
	temp = temp * temp;
	temp *= 982;

	distances[0] = (unsigned int) (temp / 200000);

	#ifdef DEBUG_JUMP_SINGLE
	uart_send_string("result: ");
	sprintf32(buffer2, "%u", (unsigned long) distances[0]);
	uart_send_string(buffer2);
	uart_send_string("\r\n");
	#endif

	menu_state = MENU_STATE_JUMP_SINGLE;

	// Display result and ask user if he/she wants to save the result.
	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("RESULT: ");
	jump_long_to_string(buffer, 8, distances[0]);
	display_4bit_write_string(buffer);
	display_4bit_write_string(" cm");
	display_4bit_set_position(0, 1);
	display_4bit_write_string("SAVE?");

	wait_for_keypad_release();

	for ( ; ; ) {
		key = keypad_scan();

		if (key == KEY_OK) {
			break;
		} else if (key == KEY_CANCEL) {
			return 1;
			break;
		} else if (key == KEY_PREV) {
			return 2;
		}
	}

	display_4bit_clear();
	display_4bit_set_position(0, 1);
	display_4bit_write_string("SAVING...");

	if (!record_store()) {
		display_4bit_set_position(0, 1);
		display_4bit_write_string("ERROR: MEM. FULL");
	} else {
		display_4bit_set_position(10, 1);
		display_4bit_write_string("OK");
	}

	delay_s(2);
	
	return 1;
}

/*
Function: test_jump_multiple

Registers:
    register0: Previous menu entry
*/
unsigned char test_jump_multiple (void)
{
	unsigned char i, j, best_result, run_test;
	unsigned long temp;
	keys_t key;
	unsigned char buffer[8];
	unsigned long air_times[MAX_SENSORS];
	unsigned long ground_times[MAX_SENSORS];

	#ifdef DEBUG_JUMP_MULTIPLE
	unsigned char buffer2[16];
	#endif

	menu_state = MENU_STATE_JUMP_MULTIPLE;

	wait_for_keypad_release();

	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("JUMP! PRESS OK");
	display_4bit_set_position(0, 1);
	display_4bit_write_string("WHEN FINISHED");	

	// Store the current time.
	fm3164_get_date(&global_record.date);

	#ifdef DEBUG_JUMP_MULTIPLE
	uart_send_string("Waiting for user to land.\r\n");
	#endif
	// Wait for the user to land.
	while (portb.RB1) {
		key = keypad_scan();
		if (key == KEY_CANCEL) { return 1; }
		else if (key == KEY_PREV) { return 2; }
	}
	counter_ms = 0;
	#ifdef DEBUG_JUMP_MULTIPLE
	uart_send_string("User has landed, counter zeroed.\r\n");
	#endif

	display_4bit_clear();
	display_4bit_set_position(2, 0);
	display_4bit_write_string("OK TO FINISH");

	run_test = 1;
	for (i = 0; i < MAX_SENSORS && run_test; ++i) {
	
		// Wait for the user to jump.

		display_4bit_set_position(0, 1);
		display_4bit_write_string("ON MAT");

		display_4bit_set_position(7, 1);
		display_4bit_write_string("JUMPS: ");
		sprintf(buffer, "%u", i);
		display_4bit_write_string(buffer);	

		#ifdef DEBUG_JUMP_MULTIPLE
		uart_send_string("Waiting for user to perform jump #");
		uitoa_dec(buffer2, i, 2);
		buffer2[2] = '\0';
		uart_send_string(buffer2);
		uart_send_string("\r\n");
		#endif

		while (!portb.RB1 && run_test) {
			key = keypad_scan();
			if (key == KEY_CANCEL) { return 1; }
			else if (key == KEY_PREV) { return 2; }
			else if (key == KEY_OK) { run_test = 0; }
		}

		// Store the ground time if we not on the first iteration.
		ground_times[i] = counter_ms;
		counter_ms = 0;

		display_4bit_set_position(0, 1);
		display_4bit_write_string("IN AIR");

		#ifdef DEBUG_JUMP_MULTIPLE
		uart_send_string("User jumped up again after ");
		sprintf32(buffer2, "%u ms on mat", ground_times[i]);
		uart_send_string(buffer2);
		uart_send_string("\r\n");
		#endif

		// Wait for user to land, abort on timeout or 'ok'.
		while (portb.RB1 && run_test && (counter_ms < 2000)) {
			key = keypad_scan();
			if (key == KEY_CANCEL || key == KEY_PREV) { return 1; }
			else if (key == KEY_OK) { run_test = 0; }
		}

		// Abort if results are ridiculus (done again because break(2); ns)
		if (counter_ms >= 2000) { break; }

		air_times[i] = counter_ms;
		counter_ms = 0;

		#ifdef DEBUG_JUMP_MULTIPLE
		uart_send_string("User landed on mat");
		sprintf32(buffer2, "%u ms in air", air_times[i]);
		uart_send_string(buffer2);
		uart_send_string("\r\n");
		#endif

	}

	// If the user aborted prematurely, subtract one from result count.
	if (!run_test) { --i; }

	// Sound the alarm to indicate last measuring.
	portc.RC2 = 1;
	delay_ms(NOTICE_BEEP_DELAY);
	portc.RC2 = 0;

	// Convert air times to jump heights.
	best_result = 0;
	for (j = 0; j < i; ++j) {
		temp = (unsigned long) (air_times[j] / 2);
		temp = temp * temp;
		temp *= 982;
		temp = (unsigned long) (temp / 200000);


		#ifdef DEBUG_JUMP_MULTIPLE
		uart_send_string("airtime: ");
		sprintf32(buffer2, "%u", air_times[j]);
		uart_send_string(buffer2);
		uart_send_string("\r\n");

		uart_send_string("result: ");
		jump_long_to_string(buffer2, 16, temp);
		uart_send_string(buffer2);
		uart_send_string("\r\n");

		uart_send_string("grtime: ");
		sprintf32(buffer2, "%u", ground_times[j]);
		uart_send_string(buffer2);
		uart_send_string("\r\n");
		#endif

		distances[j] = (unsigned int) temp;
		results[j] = ground_times[j];

		if (air_times[j] > air_times[best_result]) {
			best_result = j;
		}
	}

	global_record.num_times = i;
	global_record.type = TEST_TYPE_JUMP_MULTIPLE;
	global_record.player_id = register3;

	if (i == 0) {

		// Insult the user for not juming.
		display_4bit_clear();
		display_4bit_set_position(0, 0);
		display_4bit_write_string("NO RESULTS");
		display_4bit_set_position(0, 1);
		display_4bit_write_string("PRESS OK");

		wait_for_keypad_release();
	
		for ( ; ; ) {
			key = keypad_scan();
	
			if (key == KEY_OK) {
				break;
			} else if (key == KEY_CANCEL) {
				return 0;
				break;
			}
		}

	} else {
	// Display result and ask user if he/she wants to save the result.
	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("RESULT: ");
	jump_long_to_string(buffer, 8, (unsigned long) distances[best_result]);
	display_4bit_write_string(buffer);
	display_4bit_write_string(" cm");
	display_4bit_set_position(0, 1);
	display_4bit_write_string("SAVE?");

	display_4bit_set_position(10, 1);
	uitoa_dec(buffer, best_result + 1, 2);
	buffer[2] = '/';
	uitoa_dec(buffer + 3, j, 2);
	buffer[5] = '\0';
	display_4bit_write_string(buffer);

	wait_for_keypad_release();

	for ( ; ; ) {
		key = keypad_scan();

		if (key == KEY_OK) {
			break;
		} else if (key == KEY_CANCEL) {
			return 1;
			break;
		}
	}

	display_4bit_clear();
	display_4bit_set_position(0, 1);
	display_4bit_write_string("SAVING...");

	if (!record_store()) {
		display_4bit_set_position(0, 1);
		display_4bit_write_string("ERROR: MEM. FULL");
	} else {
		display_4bit_set_position(10, 1);
		display_4bit_write_string("OK");
	}

	delay_s(2);
	}
	
	return 1;
}

/*
Function: test_sprint_sensor

Description: 

Return value: Returns nonzero if testing should continue, zero if not.
*/
unsigned char test_sprint_sensor (void)
{	
	unsigned char sensor;
	fm3164_date_t timestamp;
	keys_t key;
	unsigned char buffer[10];

	sensor = 0;
	global_record.type = TEST_TYPE_SPRINT_SENSOR;

	display_4bit_clear();

	display_4bit_set_position(0, 0);
	display_4bit_write_string("RUN!");

	display_4bit_set_position(10, 0);
	display_4bit_write_string("--");
	display_4bit_set_position(12, 0);
	display_4bit_write_string("/");
	display_4bit_set_position(13, 0);
	display_4bit_write_integer(register1, 2);

	wait_for_keypad_release();

	// Store the current time.
	fm3164_get_date(&global_record.date);
	global_record.player_id = register3;

	// Block here until first sensor signal comes along.
	while (1) {
		key = keypad_scan();
		if (!portb.RB0) { break; }
		if (key == KEY_CANCEL) { return 0; }
		else if (key == KEY_PREV) { return 2; }
	}
	counter_ms = 0;
	while (!portb.RB0);
	display_4bit_set_position(10, 0);
	display_4bit_write_integer(sensor + 1, 2);

	// A short beep when we reach the goal. TODO: should we?
	portc.RC2 = 1;
	delay_ms(TIMESTAMP_BEEP_DELAY);
	portc.RC2 = 0;

	for (sensor = 0; sensor < register1; ++sensor) {

		// Update lap counter
		display_4bit_set_position(10, 0);
		display_4bit_write_integer(sensor + 1, 2);
		display_4bit_set_position(12, 0);
		display_4bit_write_string("/");
		display_4bit_set_position(13, 0);
		display_4bit_write_integer(register1, 2);

		while (1) {
			key = keypad_scan();
			if (!portb.RB0) { break; }
			if (key == KEY_CANCEL) { return 0; }
			else if (key == KEY_PREV) { return 2; }
			realtime_update(counter_ms, 0, 1);
		}
	
		results[sensor] = counter_ms;

		while (!portb.RB0);

		// A short beep when we reach the goal. TODO: should we?
		portc.RC2 = 1;
		delay_ms(TIMESTAMP_BEEP_DELAY);
		portc.RC2 = 0;
		//counter_ms += TIMESTAMP_BEEP_DELAY; // Add the 'delayed' time to the counter.

	}

	// The actual number of recorded results should alaways be stored.
	register2 = sensor;
	global_record.num_times = register2;

	// Display result and ask user if he/she wants to save the result.
	menu_state = MENU_STATE_SPRINT_SENSOR;

	// Display result and ask user if he/she wants to save the result.

	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("RESULT: ");
	time_long_to_string(buffer, 10, results[register2 - 1]);
	display_4bit_write_string(buffer);
	display_4bit_write_string(" s");
	display_4bit_set_position(0, 1);
	display_4bit_write_string("SAVE?");

	wait_for_keypad_release();

	for ( ; ; ) {
		key = keypad_scan();

		if (key == KEY_OK) {
			break;
		} else if (key == KEY_CANCEL) {
			return 0;
			break;
		} else if (key == KEY_PREV) {
			return 2;
		}
	}

	display_4bit_set_position(0, 1);
	display_4bit_write_string("SAVING...");

	if (!record_store()) {
		display_4bit_set_position(0, 1);
		display_4bit_write_string("ERROR: MEM. FULL");
	} else {
		display_4bit_set_position(10, 1);
		display_4bit_write_string("OK");
	}

	delay_s(2);

	return 1;
}

/*
Function: test_sprint_signal

Description: Sprint test where the athlete starts when another user presses the
    start key, exciting a beep from the machine.

Return value: Nonzero if a test

Status: Implemented, tested.
*/
unsigned char test_sprint_signal (void)
{
	unsigned char sensor;
	fm3164_date_t timestamp;
	keys_t key;
	unsigned char buffer[10];

	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("PRESS OK");
	display_4bit_set_position(0, 1);
	display_4bit_write_string("TO START");

	// Wait for the user to press next or ok before starting.
	wait_for_keypad_release();
	while (1) {
		if ((key = keypad_scan()) == KEY_OK) {
			break;
		} else if (key == KEY_CANCEL) {
			menu_state = MENU_STATE_SPRINT_SIGNAL;
			return 0;
		} else if (key == KEY_PREV) {
			return 2;
		}
	}

	wait_for_keypad_release();

	// Store the current time.
	fm3164_get_date(&global_record.date);

	global_record.type = TEST_TYPE_SPRINT_SIGNAL;

	sensor = 0;
	display_4bit_clear();
	display_4bit_set_position(10, 0);
	display_4bit_write_string("--");
	display_4bit_set_position(12, 0);
	display_4bit_write_string("/");
	display_4bit_set_position(13, 0);
	display_4bit_write_integer(register1, 2);


	display_4bit_set_position(0, 0);
	display_4bit_write_string("RUN!");

	// Sound the alarm and start sprinting.
	portc.RC2 = 1;
	counter_ms = 0;
	display_4bit_set_position(10, 0);
	display_4bit_write_integer(sensor + 1, 2);
	delay_ms(START_BEEP_DELAY);
	portc.RC2 = 0;

	for (sensor = 0; sensor < register1; ++sensor) {

		// Update lap counter
		display_4bit_set_position(10, 0);
		display_4bit_write_integer(sensor + 1, 2);
		display_4bit_set_position(12, 0);
		display_4bit_write_string("/");
		display_4bit_set_position(13, 0);
		display_4bit_write_integer(register1, 2);

		while (1) {
			key = keypad_scan();
			if (!portb.RB0) { break; }
			if (key == KEY_CANCEL) { return 0; }
			else if (key == KEY_PREV) { return 2; }
			realtime_update(counter_ms, 0, 1);
		}
		results[sensor] = counter_ms;

		// A short beep when we reach the goal. TODO: should we?
		portc.RC2 = 1;
		delay_ms(TIMESTAMP_BEEP_DELAY);
		portc.RC2 = 0;
		//counter_ms += TIMESTAMP_BEEP_DELAY; // Add the 'delayed' time to the counter.

	}

	// The actual number of recorded results should alaways be stored.
	register2 = sensor;
	global_record.num_times = register2;

	// Display result and ask user if he/she wants to save the result.
	menu_state = MENU_STATE_SPRINT_SIGNAL;
	
	// Display result and ask user if he/she wants to save the result.

	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("RESULT: ");
	time_long_to_string(buffer, 10, results[register2 - 1]);
	display_4bit_write_string(buffer);
	display_4bit_write_string(" s");
	display_4bit_set_position(0, 1);
	display_4bit_write_string("SAVE?");

	wait_for_keypad_release();

	for ( ; ; ) {
		key = keypad_scan();

		if (key == KEY_OK) {
			break;
		} else if (key == KEY_CANCEL) {
			return 0;
			break;
		} else if (key == KEY_PREV) {
			return 2;
		}
	}

	display_4bit_set_position(0, 1);
	display_4bit_write_string("SAVING...");

	if (!record_store()) {
		display_4bit_set_position(0, 1);
		display_4bit_write_string("ERROR: MEM. FULL");
	} else {
		display_4bit_set_position(10, 1);
		display_4bit_write_string("OK");
	}

	delay_s(2);

	return 1;
}

/*
Function: test_sprint_random

Description: Executes a test that a single person can execute if
    needed. Once the start button is pressed a random amount of time
    elapses before a signal indicates start.

Return value: Nonzero if the test should be repeated. Zero if shouldnt.

Registers:
    r0: menu
    r1: sensorcount
    r2: min. sensors
    r3: player id

Status: Implemented and tested.
*/
unsigned char test_sprint_random (void)
{
	unsigned char sensor;
	unsigned short random;
	fm3164_date_t timestamp;
	keys_t key;
	unsigned char buffer[10];

	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("PRESS OK");
	display_4bit_set_position(0, 1);
	display_4bit_write_string("TO START");

	// Wait for the user to press next or ok before starting.
	// Generate a random delay between 3 and 8 seconds.
	wait_for_keypad_release();
	random = 3;
	while (1) {
		if ((key = keypad_scan()) == KEY_OK) {
			break;
		} else if (key == KEY_CANCEL) {
			menu_state = MENU_STATE_SPRINT_RANDOM;
			return 0;
		} else if (key == KEY_PREV) {
			return 2;
		}

		// Reset so that we loop forever between 3 and 8.
		if (random++ >= 5) { random = 3; }
	}

	wait_for_keypad_release();
	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("WAIT...");

	// Count down the random delay time.
	delay_s(random);

	// Store the current time.
	fm3164_get_date(&global_record.date);

	sensor = 0;
	global_record.type = TEST_TYPE_SPRINT_RANDOM;
	
	display_4bit_clear();
	display_4bit_set_position(10, 0);
	display_4bit_write_string("--");
	display_4bit_set_position(12, 0);
	display_4bit_write_string("/");
	display_4bit_set_position(13, 0);
	display_4bit_write_integer(register1, 2);

	display_4bit_set_position(0, 0);
	display_4bit_write_string("RUN!");

	// Sound the alarm and start sprinting.
	portc.RC2 = 1;
	counter_ms = 0;
	delay_ms(START_BEEP_DELAY);
	portc.RC2 = 0;

	for (sensor = 0; sensor < register1; ++sensor) {

		// Update lap counter
		display_4bit_set_position(10, 0);
		display_4bit_write_integer(sensor + 1, 2);
		display_4bit_set_position(12, 0);
		display_4bit_write_string("/");
		display_4bit_set_position(13, 0);
		display_4bit_write_integer(register1, 2);

		while (1) {
			key = keypad_scan();
			if (!portb.RB0) { break; }
			if (key == KEY_CANCEL) { return 0; }
			else if (key == KEY_PREV) { return 2; }
			realtime_update(counter_ms, 0, 1);
		}
		results[sensor] = counter_ms;

		// A short beep when we reach the goal. TODO: should we?
		portc.RC2 = 1;
		delay_ms(TIMESTAMP_BEEP_DELAY);
		portc.RC2 = 0;

	}

	// The actual number of recorded results should alaways be stored.
	register2 = sensor;
	global_record.num_times = register2;

	// Display result and ask user if he/she wants to save the result.
	menu_state = MENU_STATE_SPRINT_RANDOM;
	
	// Display result and ask user if he/she wants to save the result.
	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("RESULT: ");
	time_long_to_string(buffer, 10, results[register2 - 1]);
	display_4bit_write_string(buffer);
	display_4bit_write_string(" s");
	display_4bit_set_position(0, 1);
	display_4bit_write_string("SAVE?");

	wait_for_keypad_release();

	for ( ; ; ) {
		key = keypad_scan();

		if (key == KEY_OK) {
			break;
		} else if (key == KEY_CANCEL) {
			return 0;
			break;
		} else if (key == KEY_PREV) {
			return 2;
		}
	}

	display_4bit_set_position(0, 1);
	display_4bit_write_string("SAVING...");

	if (!record_store()) {
		display_4bit_set_position(0, 1);
		display_4bit_write_string("ERROR: MEM. FULL");
	} else {
		display_4bit_set_position(10, 1);
		display_4bit_write_string("OK");
	}

	delay_s(2);

	return 1;
}

/*
Function: test_sprint_switch

Description: Executes a test where the sprinter starts with the
    foot pedal and runs past a number of radio-connected IR-sensors.

Registers:
    register0: The number of sensors used.
    register1: The player ID.
    register2: The number of recorded results.
    results: The results.
*/
unsigned char test_sprint_switch (void)
{
	unsigned char sensor;
	fm3164_date_t timestamp;
	keys_t key;
	unsigned char buffer[10];

	sensor = 0;
	global_record.type = TEST_TYPE_SPRINT_SWITCH;
	
	display_4bit_clear();
	display_4bit_set_position(10, 0);
	display_4bit_write_string("--");
	display_4bit_set_position(12, 0);
	display_4bit_write_string("/");
	display_4bit_set_position(13, 0);
	display_4bit_write_integer(register1, 2);

	display_4bit_set_position(0, 0);
	display_4bit_write_string("RUN!");

	// Store the current time.
	fm3164_get_date(&global_record.date);

	// Wait for keyboard release and pedal downpress.
	wait_for_keypad_release();
	while (portb.RB1) {
		key = keypad_scan();
		if (key == KEY_CANCEL) {
			menu_state = MENU_STATE_SPRINT_SWITCH;
			return 0;
		} else if (key == KEY_PREV) {
			return 2;
		}
		delay_ms(200);
	}

	// Wait for pedal release.
	while (!portb.RB1) {
		key = keypad_scan();
		if (key == KEY_CANCEL) {
			menu_state = MENU_STATE_SPRINT_SWITCH;
			return 0;
		} else if (key == KEY_PREV) {
			return 2;
		}
	}

	// Sound the alarm and start sprinting.
	portc.RC2 = 1;
	counter_ms = 0;
	delay_ms(START_BEEP_DELAY);
	portc.RC2 = 0;

	for (sensor = 0; sensor < register1; ++sensor) {

		// Update lap counter
		display_4bit_set_position(10, 0);
		display_4bit_write_integer(sensor + 1, 2);
		display_4bit_set_position(12, 0);
		display_4bit_write_string("/");
		display_4bit_set_position(13, 0);
		display_4bit_write_integer(register1, 2);

		while (1) {
			key = keypad_scan();
			if (!portb.RB0) { break; }
			if (key == KEY_CANCEL) { return 0; }
			else if (key == KEY_PREV) { return 2; }
			realtime_update(counter_ms, 0, 1);
		}
		results[sensor] = counter_ms;

		// A short beep when we reach the goal. TODO: should we?
		portc.RC2 = 1;
		delay_ms(TIMESTAMP_BEEP_DELAY);
		portc.RC2 = 0;
	}

	// The actual number of recorded results should alaways be stored.
	register2 = sensor;
	global_record.num_times = register2;

	// Display result and ask user if he/she wants to save the result.
	menu_state = MENU_STATE_SPRINT_SWITCH;
	
	// Display result and ask user if he/she wants to save the result.
	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("RESULT: ");
	time_long_to_string(buffer, 10, results[register2 - 1]);
	display_4bit_write_string(buffer);
	display_4bit_write_string(" s");
	display_4bit_set_position(0, 1);
	display_4bit_write_string("SAVE?");

	wait_for_keypad_release();

	for ( ; ; ) {
		key = keypad_scan();

		if (key == KEY_OK) {
			break;
		} else if (key == KEY_CANCEL) {
			return 0;
			break;
		} else if (key == KEY_PREV) {
			return 2;
		}
	}

	display_4bit_set_position(0, 1);
	display_4bit_write_string("SAVING...");

	if (!record_store()) {
		display_4bit_set_position(0, 1);
		display_4bit_write_string("ERROR: MEM. FULL");
	} else {
		display_4bit_set_position(10, 1);
		display_4bit_write_string("OK");
	}

	delay_s(2);
}

void menu_process (keys_t key)
{
	unsigned char ret;

	// Hack, always check for presses on the PRINT button.
	if (key == KEY_PRINT) {
		register0 = menu_state;
		transmit_records();
		menu_clear_memory();
		return;
	}

	switch (menu_state)
	{
	case MENU_STATE_SPRINT:
		switch (key)
		{
		case KEY_SUB_INC:
			menu_state = MENU_STATE_JUMP;
			break;
		
		case KEY_POP_DEC:
			menu_state = MENU_STATE_SETUP;
			break;

		case KEY_OK:
		case KEY_NEXT:
			menu_state = MENU_STATE_SPRINT_SENSOR;
			break;

		case KEY_PRINT:
			transmit_records();
			register0 = (menu_states_t) MENU_STATE_SPRINT;
			menu_clear_memory();
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_SPRINT_SENSOR:
		switch (key)
		{
		case KEY_SUB_INC:
			menu_state = MENU_STATE_SPRINT_SIGNAL;
			break;
		
		case KEY_POP_DEC:
			menu_state = MENU_STATE_SPRINT_SWITCH;
			break;

		case KEY_OK:
		case KEY_NEXT:
			register0 = MENU_STATE_SPRINT_SENSOR;
			register1 = 2; // Default value of 2 sensor.
			register2 = 2; // Minimum of 2 sensors.
			register3 = 0;
			if (setup_sprint_test(0, TEST_TYPE_SPRINT_SENSOR)) {
				while ((ret = test_sprint_sensor())) {
					if (ret == 2) { ret = 1; ++register1; }
					if (!setup_sprint_test(ret, TEST_TYPE_SPRINT_SENSOR)) { break; }
				}
			}
			break;

		case KEY_PREV:
		case KEY_CANCEL:
			menu_state = MENU_STATE_SPRINT;
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_SPRINT_SIGNAL:
		switch (key)
		{
		case KEY_SUB_INC:
			menu_state = MENU_STATE_SPRINT_RANDOM;
			break;
		
		case KEY_POP_DEC:
			menu_state = MENU_STATE_SPRINT_SENSOR;
			break;

		case KEY_OK:
		case KEY_NEXT:
			register0 = MENU_STATE_SPRINT_SIGNAL;
			register1 = 1; // One sensor is default.
			register2 = 1; // Minimum of one sensor.
			if (setup_sprint_test(0, TEST_TYPE_SPRINT_SIGNAL)) {
				while ((ret = test_sprint_signal())) {
					if (ret == 2) { ret = 1; }
					if (!setup_sprint_test(ret, TEST_TYPE_SPRINT_SIGNAL)) { break; }
				}
			}
			break;

		case KEY_PREV:
		case KEY_CANCEL:
			menu_state = MENU_STATE_SPRINT;
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_SPRINT_RANDOM:
		switch (key)
		{
		case KEY_SUB_INC:
			menu_state = MENU_STATE_SPRINT_SWITCH;
			break;
		
		case KEY_POP_DEC:
			menu_state = MENU_STATE_SPRINT_SIGNAL;
			break;

		case KEY_OK:
		case KEY_NEXT:
			register0 = MENU_STATE_SPRINT_RANDOM;
			register1 = 1; // Default to one sensor.
			register2 = 1; // Minimum of one sensor.
			if (setup_sprint_test(0, TEST_TYPE_SPRINT_RANDOM)) {
				while ((ret = test_sprint_random())) {
					if (ret == 2) { ret = 1; }
					if (!setup_sprint_test(ret, TEST_TYPE_SPRINT_RANDOM)) { break; }
				}
			}
			break;

		case KEY_PREV:
		case KEY_CANCEL:
			menu_state = MENU_STATE_SPRINT;
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_SPRINT_SWITCH:
		switch (key)
		{
		case KEY_SUB_INC:
			menu_state = MENU_STATE_SPRINT_SENSOR;
			break;
		
		case KEY_POP_DEC:
			menu_state = MENU_STATE_SPRINT_RANDOM;
			break;

		case KEY_OK:
		case KEY_NEXT:
			register0 = MENU_STATE_SPRINT_SWITCH;
			register1 = 1; // Default to one sensor pair.
			register2 = 1; // Minimum of one sensor.
			if (setup_sprint_test(0, TEST_TYPE_SPRINT_SWITCH)) {
				while ((ret = test_sprint_switch())) {
					if (ret == 2) { ret = 1; }
					if (!setup_sprint_test(ret, TEST_TYPE_SPRINT_SWITCH)) { break; }
				}
			}
			break;

		case KEY_PREV:
		case KEY_CANCEL:
			menu_state = MENU_STATE_SPRINT;
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_JUMP:
		switch (key)
		{
		case KEY_SUB_INC:
			menu_state = MENU_STATE_ENDURANCE;
			break;
		
		case KEY_POP_DEC:
			menu_state = MENU_STATE_SPRINT;
			break;

		case KEY_OK:
		case KEY_NEXT:
			menu_state = MENU_STATE_JUMP_SINGLE;
			break;

		case KEY_PRINT:
			transmit_records();
			register0 = (menu_states_t) MENU_STATE_JUMP;
			menu_clear_memory();
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_JUMP_SINGLE:
		switch (key)
		{
		case KEY_SUB_INC:
			menu_state = MENU_STATE_JUMP_MULTIPLE;
			break;
		
		case KEY_POP_DEC:
			menu_state = MENU_STATE_JUMP_MULTIPLE;
			break;

		case KEY_PREV:
		case KEY_CANCEL:
			menu_state = MENU_STATE_JUMP;
			break;

		case KEY_OK:
		case KEY_NEXT:
			register0 = MENU_STATE_JUMP_SINGLE; // Return to jump submenu for PREV.
			register1 = MENU_STATE_JUMP; // Return to main menu for CANCEL.
			register3 = 0; // Default player id.
			if (setup_jump_test(0)) {
				while ((ret = test_jump_single())) {
					if (ret == 2) { ret = 0; }
					if (!setup_jump_test(ret)) { break; }
				}
			}
			break;	

		default:
			break;
		};
		break;

	case MENU_STATE_JUMP_MULTIPLE:
		switch (key)
		{
		case KEY_SUB_INC:
		case KEY_POP_DEC:
			menu_state = MENU_STATE_JUMP_SINGLE;
			break;
		
		case KEY_PREV:
		case KEY_CANCEL:
			menu_state = MENU_STATE_JUMP;
			break;

		case KEY_OK:
		case KEY_NEXT:
			// A press on PREV takes the user to the JUMP submenu.
			register0 = (menu_states_t) MENU_STATE_JUMP_MULTIPLE;
			// A press on CANCEL takes user to the main menu, JUMP item.
			register1 = (menu_states_t) MENU_STATE_JUMP;
			if (setup_jump_test(0)) {
				while ((ret = test_jump_multiple())) {
					if (ret == 2) { ret = 0; }
					if (!setup_jump_test(ret)) { break; }
				}
			}
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_ENDURANCE:
		switch (key)
		{
		case KEY_SUB_INC:
			menu_state = MENU_STATE_RESULTS;
			break;
		
		case KEY_POP_DEC:
			menu_state = MENU_STATE_JUMP;
			break;

		case KEY_OK:
		case KEY_NEXT:
			menu_state = MENU_STATE_ENDURANCE_MULTI;
			break;

		case KEY_PRINT:
			transmit_records();
			register0 = (menu_states_t) MENU_STATE_ENDURANCE;
			menu_clear_memory();
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_ENDURANCE_BEEP:
		switch (key)
		{
		case KEY_SUB_INC:
		case KEY_POP_DEC:
			menu_state = MENU_STATE_ENDURANCE_MULTI;
			break;

		case KEY_OK:
		case KEY_NEXT:
			while (test_endurance_beep());
			break;

		case KEY_CANCEL:
		case KEY_PREV:
			menu_state = MENU_STATE_ENDURANCE;
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_ENDURANCE_MULTI:
		switch (key)
		{
		case KEY_SUB_INC:	
		case KEY_POP_DEC:
			menu_state = MENU_STATE_ENDURANCE_BEEP;
			break;

		case KEY_OK:
		case KEY_NEXT:
			register0 = MENU_STATE_ENDURANCE_MULTI; // Prev button
			register1 = MENU_STATE_ENDURANCE; // Cancel button
			register3 = 0;
			if (setup_endurance_test(0)) {
				while ((ret = test_endurance_multi())) {
					if (ret == 2) { ret = 1; }
					if (!setup_endurance_test(ret)) { break; }
				}
			}
			break;

		case KEY_PREV:
		case KEY_CANCEL:
			menu_state = MENU_STATE_ENDURANCE;
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_RESULTS:
		switch (key)
		{
		case KEY_SUB_INC:
			menu_state = MENU_STATE_SETUP;
			break;
		
		case KEY_POP_DEC:
			menu_state = MENU_STATE_ENDURANCE;
			break;

		case KEY_OK:
		case KEY_NEXT:
			menu_state = MENU_STATE_RESULTS_SPRINT;
			break;

		case KEY_PRINT:
			transmit_records();
			register0 = (menu_states_t) MENU_STATE_RESULTS;
			menu_clear_memory();
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_RESULTS_SPRINT:
		switch (key)
		{
		case KEY_SUB_INC:
			menu_state = MENU_STATE_RESULTS_JUMP;
			break;

		case KEY_POP_DEC:
			menu_state = MENU_STATE_RESULTS_ENDURANCE;
			break;
		
		case KEY_PREV:
		case KEY_CANCEL:
			menu_state = MENU_STATE_RESULTS;
			break;

		case KEY_OK:
		case KEY_NEXT:
			menu_results_sprint();
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_RESULTS_JUMP:
		switch (key)
		{
		case KEY_SUB_INC:
			menu_state = MENU_STATE_RESULTS_ENDURANCE;
			break;
		
		case KEY_POP_DEC:
			menu_state = MENU_STATE_RESULTS_SPRINT;
			break;

		case KEY_OK:
		case KEY_NEXT:
			menu_results_jump();
			break;

		case KEY_PREV:
		case KEY_CANCEL:
			menu_state = MENU_STATE_RESULTS;
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_RESULTS_ENDURANCE:
		switch (key)
		{
		case KEY_SUB_INC:
			menu_state = MENU_STATE_RESULTS_SPRINT;
			break;
		
		case KEY_POP_DEC:
			menu_state = MENU_STATE_RESULTS_JUMP;
			break;

		case KEY_OK:
		case KEY_NEXT:
			menu_results_endurance();
			break;

		case KEY_PREV:
		case KEY_CANCEL:
			menu_state = MENU_STATE_RESULTS;
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_SETUP:
		switch (key)
		{
		case KEY_SUB_INC:
			menu_state = MENU_STATE_SPRINT;
			break;
		
		case KEY_POP_DEC:
			menu_state = MENU_STATE_RESULTS;
			break;

		case KEY_OK:
		case KEY_NEXT:
			menu_state = MENU_STATE_SETUP_DATE;
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_SETUP_DATE:
		switch (key)
		{
		case KEY_SUB_INC:
			menu_state = MENU_STATE_SETUP_MEMORY;
			break;

		case KEY_POP_DEC:
			menu_state = MENU_STATE_SETUP_SOFTWARE;
			break;

		case KEY_OK:
		case KEY_NEXT:
			menu_setup_date();
			break;

		case KEY_CANCEL:
		case KEY_PREV:
			menu_state = MENU_STATE_SETUP;
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_SETUP_MEMORY:
		switch (key)
		{
		case KEY_SUB_INC:
			menu_state = MENU_STATE_SETUP_SOFTWARE;
			break;

		case KEY_POP_DEC:
			menu_state = MENU_STATE_SETUP_DATE;
			break;

		case KEY_OK:
		case KEY_NEXT:
			menu_state = MENU_STATE_SETUP_MEMORY_STATUS;
			break;

		case KEY_CANCEL:
		case KEY_PREV:
			menu_state = MENU_STATE_SETUP;
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_SETUP_MEMORY_STATUS:
		switch (key)
		{
		case KEY_PREV:
			menu_state = MENU_STATE_SETUP_MEMORY;
			break;

		case KEY_SUB_INC:
			menu_state = MENU_STATE_SETUP_MEMORY_CLEARPLAYER;
			break;

		case KEY_POP_DEC:
			menu_state = MENU_STATE_SETUP_MEMORY_CLEAR;
			break;

		case KEY_NEXT:
		case KEY_OK:
			menu_memory_status();
			break;

		case KEY_CANCEL:
			menu_state = MENU_STATE_SETUP;
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_SETUP_MEMORY_CLEARPLAYER:
		switch (key)
		{
		case KEY_PREV:
			menu_state = MENU_STATE_SETUP_MEMORY;
			break;

		case KEY_SUB_INC:
			menu_state = MENU_STATE_SETUP_MEMORY_CLEAR;
			break;

		case KEY_POP_DEC:
			menu_state = MENU_STATE_SETUP_MEMORY_STATUS;
			break;

		case KEY_NEXT:
		case KEY_OK:
			// A press on PREV takes the user back to the setup/memory menu with clearplayer
			// selected as the current item.
			register0 = (menu_states_t) MENU_STATE_SETUP_MEMORY_CLEARPLAYER;
			// A press on CANCEL takes user to the main menu, JUMP item.
			register1 = (menu_states_t) MENU_STATE_SETUP_MEMORY;
			register3 = 0; // Player ID to clear (modified by setup_jump_test).
			if (setup_jump_test(0)) {
				while (menu_clear_player()) {
					//register1 = (menu_states_t) MENU_STATE_SETUP_MEMORY;
					if (!setup_jump_test(1)) { break; }
					//register1 = (menu_states_t) MENU_STATE_SETUP_CLEARPLAYER;
				}
			}
			break;

		case KEY_CANCEL:
			menu_state = MENU_STATE_SETUP;
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_SETUP_MEMORY_CLEAR:
		switch (key)
		{
		case KEY_PREV:
			menu_state = MENU_STATE_SETUP_MEMORY;
			break;

		case KEY_SUB_INC:
			menu_state = MENU_STATE_SETUP_MEMORY_STATUS;
			break;

		case KEY_POP_DEC:
			menu_state = MENU_STATE_SETUP_MEMORY_CLEARPLAYER;
			break;

		case KEY_CANCEL:
			menu_state = MENU_STATE_SETUP;
			break;

		case KEY_NEXT:
		case KEY_OK:
			register0 = (menu_states_t) MENU_STATE_SETUP_MEMORY_CLEAR;
			menu_clear_memory();
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_SETUP_SOFTWARE:
		switch (key)
		{
		case KEY_POP_DEC:
			menu_state = MENU_STATE_SETUP_MEMORY;
			break;

		case KEY_SUB_INC:
			menu_state = MENU_STATE_SETUP_DATE;
			break;

		case KEY_CANCEL:
		case KEY_PREV:
			menu_state = MENU_STATE_SETUP;
			break;

		case KEY_OK:
		case KEY_NEXT:
			menu_state = MENU_STATE_SETUP_SOFTWARE_DISPLAY;
			break;

		default:
			break;
		};
		break;

	case MENU_STATE_SETUP_SOFTWARE_DISPLAY:
		switch (key)
		{
		case KEY_PREV:
		case KEY_OK:
			menu_state = MENU_STATE_SETUP_SOFTWARE;
			break;

		case KEY_CANCEL:
			menu_state = MENU_STATE_SETUP;
			break;

		default:
			break;
		};
		break;
	};
}

/*
Function: draw_markers
Description: Draws the menu markers on the screen.
*/
void draw_markers (void)
{
	display_4bit_set_position(0, 1);
	display_4bit_write_character('<');
	display_4bit_set_position(15, 1);
	display_4bit_write_character('>');
}

/*
Function: realtime_update
    Prints the time at the designated display location.

Parameters:
    result: The timeresult to print.
    pos_x: Horizontal display position (0-15)
    pos_y: Vertical display position (0-1)
*/
void realtime_update (unsigned long result,
	unsigned char pos_x, unsigned char pos_y)
{
	char buffer[13];

	display_4bit_set_position(pos_x, pos_y);
	util_time_to_string(result, buffer);
	display_4bit_write_string(buffer);
}

/*
Function: menu_draw
Description: Handles redrawing for the menu system given a specific state.
*/
void menu_draw (void)
{
	char buffer[13];

	display_4bit_clear();

	switch (menu_state)
	{
	case MENU_STATE_SPRINT:
		display_4bit_set_position(6, 0);
		display_4bit_write_string("MAIN");
		display_4bit_set_position(5, 1);
		display_4bit_write_string("SPRINT");
		draw_markers();
		break;

	case MENU_STATE_SPRINT_SENSOR:
		display_4bit_set_position(5, 0);
		display_4bit_write_string("SPRINT");
		display_4bit_set_position(2, 1);
		display_4bit_write_string("SENSOR START");
		draw_markers();
		break;

	case MENU_STATE_SPRINT_SIGNAL:
		display_4bit_set_position(5, 0);
		display_4bit_write_string("SPRINT");
		display_4bit_set_position(2, 1);
		display_4bit_write_string("SIGNAL START");
		draw_markers();
		break;

	case MENU_STATE_SPRINT_RANDOM:
		display_4bit_set_position(5, 0);
		display_4bit_write_string("SPRINT");
		display_4bit_set_position(2, 1);
		display_4bit_write_string("RANDOM START");
		draw_markers();
		break;

	case MENU_STATE_SPRINT_SWITCH:
		display_4bit_set_position(5, 0);
		display_4bit_write_string("SPRINT");
		display_4bit_set_position(2, 1);
		display_4bit_write_string("SWITCH START");
		draw_markers();
		break;

	case MENU_STATE_JUMP:
		display_4bit_set_position(6, 0);
		display_4bit_write_string("MAIN");
		display_4bit_set_position(6, 1);
		display_4bit_write_string("JUMP");
		draw_markers();
		break;

	case MENU_STATE_JUMP_SINGLE:
		display_4bit_set_position(6, 0);
		display_4bit_write_string("JUMP");
		display_4bit_set_position(5, 1);
		display_4bit_write_string("STATIC");
		draw_markers();
		break;

	case MENU_STATE_JUMP_MULTIPLE:
		display_4bit_set_position(6, 0);
		display_4bit_write_string("JUMP");
		display_4bit_set_position(5, 1);
		display_4bit_write_string("REBOUND");
		draw_markers();
		break;

	case MENU_STATE_ENDURANCE:
		display_4bit_set_position(6, 0);
		display_4bit_write_string("MAIN");
		display_4bit_set_position(4, 1);
		display_4bit_write_string("ENDURANCE");
		draw_markers();
		break;

	case MENU_STATE_ENDURANCE_BEEP:
		display_4bit_set_position(4, 0);
		display_4bit_write_string("ENDURANCE");
		display_4bit_set_position(4, 1);
		display_4bit_write_string("BEEP TEST");
		draw_markers();
		break;

	case MENU_STATE_ENDURANCE_MULTI:
		display_4bit_set_position(4, 0);
		display_4bit_write_string("ENDURANCE");
		display_4bit_set_position(2, 1);
		display_4bit_write_string("MULTI SPRINT");
		draw_markers();
		break;

	case MENU_STATE_RESULTS:
		display_4bit_set_position(6, 0);
		display_4bit_write_string("MAIN");
		display_4bit_set_position(5, 1);
		display_4bit_write_string("RESULTS");
		draw_markers();
		break;

	case MENU_STATE_RESULTS_SPRINT:
		display_4bit_set_position(5, 0);
		display_4bit_write_string("RESULTS");
		display_4bit_set_position(5, 1);
		display_4bit_write_string("SPRINT");
		draw_markers();
		break;

	case MENU_STATE_RESULTS_JUMP:
		display_4bit_set_position(5, 0);
		display_4bit_write_string("RESULTS");
		display_4bit_set_position(6, 1);
		display_4bit_write_string("JUMP");
		draw_markers();
		break;

	case MENU_STATE_RESULTS_ENDURANCE:
		display_4bit_set_position(5, 0);
		display_4bit_write_string("RESULTS");
		display_4bit_set_position(4, 1);
		display_4bit_write_string("ENDURANCE");
		draw_markers();
		break;

	case MENU_STATE_SETUP:
		display_4bit_set_position(6, 0);
		display_4bit_write_string("MAIN");
		display_4bit_set_position(5, 1);
		display_4bit_write_string("SETUP");
		draw_markers();
		break;

	case MENU_STATE_SETUP_DATE:
		display_4bit_set_position(5, 0);
		display_4bit_write_string("SETUP");
		display_4bit_set_position(3, 1);
		display_4bit_write_string("DATE, TIME");
		draw_markers();
		break;

	case MENU_STATE_SETUP_MEMORY:
		display_4bit_set_position(5, 0);
		display_4bit_write_string("SETUP");
		display_4bit_set_position(5, 1);
		display_4bit_write_string("MEMORY");
		draw_markers();
		break;

	case MENU_STATE_SETUP_MEMORY_STATUS:
		display_4bit_set_position(5, 0);
		display_4bit_write_string("MEMORY");
		display_4bit_set_position(5, 1);
		display_4bit_write_string("STATUS");
		draw_markers();
		break;

	case MENU_STATE_SETUP_MEMORY_CLEARPLAYER:
		display_4bit_set_position(5, 0);
		display_4bit_write_string("MEMORY");
		display_4bit_set_position(2, 1);
		display_4bit_write_string("CLEAR PLAYER");
		draw_markers();
		break;

	case MENU_STATE_SETUP_MEMORY_CLEAR:
		display_4bit_set_position(5, 0);
		display_4bit_write_string("MEMORY");
		display_4bit_set_position(5, 1);
		display_4bit_write_string("CLEAR");
		draw_markers();
		break;

	case MENU_STATE_SETUP_SOFTWARE:	
		display_4bit_set_position(5, 0);
		display_4bit_write_string("SETUP");
		display_4bit_set_position(4, 1);
		display_4bit_write_string("SOFTWARE");
		draw_markers();
		break;

	case MENU_STATE_SETUP_SOFTWARE_DISPLAY:
		display_4bit_set_position(2, 0);
		display_4bit_write_string("TIME IT 2008");
		display_4bit_set_position(2, 1);
		display_4bit_write_string("Version 1.00");
		draw_markers();
		break;

	case MENU_STATE_DOWNLOAD_TRANSMITTING:
		display_4bit_set_position(2, 0);
		display_4bit_write_string("TRANSMITTING");
		display_4bit_set_position(2, 1);
		display_4bit_write_string("TO COMPUTER");
		break;
	};
}

/*
Function: interrupt
Description: Standard interrupt handler, only used to increment counter_ms.
*/
void interrupt (void)
{
	static unsigned int fulhack = 0;

	// Check if we have our 1-ms interrupt flagged.
	if (pir1.TMR2IF == 1) {

		if (fulhack > 4) {
			counter_ms++;
			fulhack = 0;
		}

		++fulhack;

		pir1.TMR2IF = 0;
	}
}

/*
Function: date_modify
Description: Allows the user to input a new date into the Ramtron device.
*/
void date_modify (void)
{
	unsigned char data[6];
	unsigned char cursor = 0;
	keys_t key;

	while ((key = keypad_scan()) != KEY_CANCEL) {
		switch (key)
		{
		case KEY_0:
			data[cursor] = 0;
			break;

		case KEY_1:
			data[cursor] = 1;
			break;

		case KEY_2:
			data[cursor] = 2;
			break;

		case KEY_3:
			data[cursor] = 3;
			break;

		case KEY_4:
			data[cursor] = 4;
			break;

		case KEY_5:
			data[cursor] = 5;
			break;

		case KEY_6:
			data[cursor] = 6;
			break;

		case KEY_7:
			data[cursor] = 7;
			break;

		case KEY_8:
			data[cursor] = 8;
			break;

		case KEY_9:
			data[cursor] = 9;
			break;

		case KEY_POP_DEC:
			if (cursor > 0) { --cursor; }
			break;

		case KEY_SUB_INC:
			if (cursor < 5) { ++cursor; }
			break;

		case KEY_NEXT:
			if (data[cursor] > 0x00) { data[cursor]--; }
			break;

		case KEY_PREV:
			if (data[cursor] < 0xff) { data[cursor]++; }
			break;

		case KEY_OK:
			break;

		default:
			break;
		};
	}
}

/*
Function: record_store
Description: Writes the contents of 'global_record', 'results' and 'distances'
    to the eeprom memory.
*/
unsigned char record_store (void)
{
	unsigned int end, tmp;
	unsigned int size_of_times;

	#ifdef DEBUG_WRITE_RECORDS
	unsigned char buffer[40];
	#endif

	// Calculate the size, in bytes, of the entire record.
	size_of_times = (sizeof(unsigned long) + sizeof(unsigned int)) * global_record.num_times;

	// Find the location in the EEPROM for the last record.
	// This information is stored at address 0x00 in the
	// EEPROM and is an 16 bit unsigned integer.
	fm3164_read_memory (0x0000, (unsigned char *) &end, sizeof(unsigned int));

	// Is there enough space free in the memory?
	if (FM3164_MEMORY_SIZE - end < (sizeof(record_t) + size_of_times)) { return 0; }

	// Write the record header to eeprom memory.
	fm3164_write_memory(end, (const unsigned char *) &global_record, sizeof(record_t));

	end += sizeof(record_t);

	#ifdef DEBUG_WRITE_RECORDS
	sprintf(buffer, "Records: %u\r\n", global_record.num_times);
	uart_send_string(buffer);
	#endif

	// Write the time records and distances to eeprom memory.
	for (tmp = 0; tmp < global_record.num_times; ++tmp) {
		
		fm3164_write_memory(end, (const unsigned char *) &results[tmp],
			sizeof(unsigned long));
		end += sizeof(unsigned long);

		#ifdef DEBUG_WRITE_RECORDS
		sprintf32(buffer, "Result: %ul, distance: %u\r\n", results[tmp], distances[tmp]);
		#endif

		fm3164_write_memory(end, (const unsigned char *) &distances[tmp],
			sizeof(unsigned int));
		end += sizeof(unsigned int);
	}

	// Update the end marker record.
	fm3164_write_memory(0x0000, (const unsigned char *) &end, sizeof(unsigned int));

	return 1;
}

/*
Function: record_retrieve
Description: Reads the address specified by 'current_record' into 'global_record' and 'results'.
Return value: The size of the currently loaded record, in bytes.
*/
unsigned int record_retrieve (void)
{
	unsigned int temp, size;

	#ifdef DEBUG_EEPROM
	unsigned char buffer[32];
	#endif

	size = 0;

	#ifdef DEBUG_EEPROM
	sprintf(buffer, "rr, cur_r: 0x%X\r\n", current_record);
	uart_send_string(buffer);
	#endif

	fm3164_read_memory(current_record, (unsigned char *) &global_record, sizeof(record_t));
	current_record += sizeof(record_t);
	size += sizeof(record_t);

	// Read the time records from eeprom memory.
	for (temp = 0; temp < global_record.num_times; ++temp) {
		
		fm3164_read_memory(current_record, (unsigned char *) &results[temp],
			sizeof(unsigned long));
		current_record += sizeof(unsigned long);
		size += sizeof(unsigned long);

		fm3164_read_memory(current_record, (unsigned char *) &distances[temp],
			sizeof(unsigned int));
		current_record += sizeof(unsigned int);
		size += sizeof(unsigned int);
	}

	current_record_size = size;

	return size;
}

/*
Function: record_retrieve_previous
Description: Supply start address of current record, this function will
    load the record immidately before the address you specified and
    return the size of that record in bytes.
*/
unsigned int record_retrieve_previous (unsigned int old_address)
{
	unsigned int location, record_size;
	unsigned char buffer[20];

	location = 0x0002;
	record_size = 0;
	current_record = 0x0002;

	// Search memory until we find a record whose size adds up
	// to the memory location supplied as the old address.
	while (location < old_address) {

		record_size = record_retrieve();

		location += record_size;

		#ifdef DEBUG_EEPROM
		sprintf(buffer, "size: %u\r\n", record_size);
		uart_send_string(buffer);
		sprintf(buffer, "loc: %u", location);
		uart_send_string(buffer);
		sprintf(buffer, "old: %u", old_address);
		uart_send_string(buffer);
		#endif
	}

	return record_size;
}

/*
Function: transmit_records
Description: Serializes the EEPROM memory contents and sends them to the
    host computer using the serial port. The protocol used is TRTP 1.0.
    See separate documentation for details of the TRTP protocol.
*/
void transmit_records (void)
{
	unsigned int end;
	unsigned char buffer[16];
	unsigned char i, value;
	unsigned char hours, minutes, seconds;
	unsigned long temp;

	// Send a questio mark in order to indicate start of transmission.
	uart_send_character('?');

	// The first 2 octets points to the end of the data block.
	fm3164_read_memory(0x0000, (unsigned char *) &end, 2);

	current_record = 0x0002;
	while (current_record < end) {
		
		// Load the next record into 'global_record'.
		record_retrieve();

		// Serialize and send the player id.
		memset(buffer, '\0', sizeof(buffer));
		sprintf(buffer, "%u;", (unsigned int) global_record.player_id);
		uart_send_string(buffer);

/*
		// COMMENTED OUT: Test ID are not used any longer.

		// Serialize and send the test id.
		memset(buffer, '\0', sizeof(buffer));
		sprintf(buffer, "%u;", (unsigned int) global_record.test_id);
		uart_send_string(buffer);
*/

		// Serialize and send the test type.
		switch (global_record.type)
		{
		case TEST_TYPE_JUMP_SINGLE:
			buffer[0] = 'J'; buffer[1] = 'S';
			break;
		case TEST_TYPE_JUMP_MULTIPLE:
			buffer[0] = 'J'; buffer[1] = 'M';
			break;
		case TEST_TYPE_SPRINT_SENSOR:
			buffer[0] = 'S'; buffer[1] = 'S';
			break;
		case TEST_TYPE_SPRINT_SIGNAL:
			buffer[0] = 'S'; buffer[1] = 'I';
			break;
		case TEST_TYPE_SPRINT_RANDOM:
			buffer[0] = 'S'; buffer[1] = 'R';
			break;
		case TEST_TYPE_SPRINT_SWITCH:
			buffer[0] = 'S'; buffer[1] = 'W';
			break;
		case TEST_TYPE_ENDURANCE_BEEP:
			buffer[0] = 'B'; buffer[1] = 'T';
			break;
		case TEST_TYPE_ENDURANCE_MULTI:
			buffer[0] = 'M'; buffer[1] = 'S';
			break;
		};
		buffer[2] = ';'; buffer[3] = '\0';
		uart_send_string(buffer);

		// Serialize and send the date.
		uitoa_dec(buffer, global_record.date.year + 2000, 4);
		uitoa_dec(buffer + 4, global_record.date.month, 2);
		uitoa_dec(buffer + 6, global_record.date.day, 2);
		uitoa_dec(buffer + 8, global_record.date.hours, 2);
		uitoa_dec(buffer + 10, global_record.date.minutes, 2);
		uitoa_dec(buffer + 12, global_record.date.seconds, 2);
		buffer[14] = ';'; buffer[15] = '\0';
		uart_send_string(buffer);

		if (global_record.num_times > MAX_SENSORS) {
			uart_send_string("$@\r\nInternal error: Invalid result count.\r\n");
			return;
		}

		// Serialize and send each timing result.
		for (i = 0; i < global_record.num_times; ++i) {
		
			memset(buffer, '\0', sizeof(buffer));
			sprintf32(buffer, "%u;", results[i]);
			uart_send_string(buffer);
			
			// Send the distance.
			memset(buffer, '\0', sizeof(buffer));
			sprintf32(buffer, "%u;", distances[i]);
			uart_send_string(buffer);
		}

		// Send the record terminator (a dollar sign).
		uart_send_character('$');

	}

	// Send the transmission terminator (a 'kanelbulle').
	uart_send_character('@');
	uart_send_character('\r');
	uart_send_character('\n');
}



/*
Function: init_eeprom

Description: Reinitalizes the EEPROM file storage to empty.
*/
void init_eeprom (void)
{
	unsigned int end;

	// Reset the end of marker record.
	end = 0x0002;
	fm3164_write_memory(0x0000, (const unsigned char *) &end, 2);

	current_record = 0x0002;
}

/*
Function: setup_sprint_test
Description: This function allows the user to setup the parameters for a sprint test.
Parameters:
    reuse: If not zero, only the player id will be modified.

    registers[0]: menu to return to.
    registers[1]: sensor count
    registers[2]: minimal sensor count
    registers[3]: player id
*/
unsigned char setup_sprint_test (unsigned char reuse, test_types_t test_type)
{
	keys_t key;
	unsigned char value, i;
	unsigned int distance;

	// This variable is used to track where any direct number
	// entry happened in the input string.
	unsigned char input_position;

setup_sprint_1:
	if (!reuse) {
		// Allow the user to input the number of sensors into register1.

		// Make sure we get the minimum sensor count right.
		if (register1 < register2) { register1 = register2; }

		display_4bit_clear();
	
		display_4bit_set_position(0, 0);
		display_4bit_write_string("NUMBER OF");
		display_4bit_set_position(0, 1);
		display_4bit_write_string("SENSORS:");
		display_4bit_set_position(9, 1);
		display_4bit_write_integer(register1, 2);
	
		input_position = 0;
		wait_for_keypad_release();
		while ((key = keypad_scan()) != KEY_OK && key != KEY_NEXT) {
			wait_for_keypad_release();
			value = 10; // This is an invalid number.
	
			switch (key)
			{
			case KEY_SUB_INC:
				if (register1 < 99) { ++register1; }
				input_position = 0;
				display_4bit_set_position(9, 1);
				display_4bit_write_integer(register1, 2);
				break;
				
			case KEY_POP_DEC:
				if (register1 > register2) { --register1; }
				input_position = 0;
				display_4bit_set_position(9, 1);
				display_4bit_write_integer(register1, 2);
				break;
		
			// This menu entry causes the config to abort and returns us to
			// to the appropiate menu entry from whence we came.
			case KEY_PREV:
				menu_state = (menu_states_t) register0;
				return 0;
				break;
		
			// Abort alltogether and return to initial state.
			case KEY_CANCEL:
				menu_state = MENU_STATE_SPRINT;
				return 0;
				break;
	
			case KEY_0:
				value = 0;
				break;
	
			case KEY_1:
				value = 1;
				break;
	
			case KEY_2:
				value = 2;
				break;
	
			case KEY_3:
				value = 3;
				break;
	
			case KEY_4:
				value = 4;
				break;
	
			case KEY_5:
				value = 5;
				break;
	
			case KEY_6:
				value = 6;
				break;
	
			case KEY_7:
				value = 7;
				break;
	
			case KEY_8:
				value = 8;
				break;
	
			case KEY_9:
				value = 9;
				break;
		
			default:
				break;
			};

			// If value have been modified someone has input a digit. Process it.
			if (value != 10) {
				if (counter_ms > 2000) { input_position = 0; register1 = 0; }
				// We must zero any value if this is the start of the keyboard
				// input sequence.
				if (input_position == 0) { register1 = 0; }
				register1 *= 10;
				register1 += value;
				//register1 += value * math_pow(10, input_position);
				if (register1 < register2) { register1 = register2; }
				else if (register1 > MAX_SENSORS) { register1 = MAX_SENSORS; }
				
				display_4bit_set_position(9, 1);
				display_4bit_write_integer(register1, 2);
	
				// We only support two digits, so reset 'input_position' if needed.
				++input_position;
				if (input_position > 1) { input_position = 0; }

				counter_ms = 0;
			}

			delay_ms(KEY_PRESS_DELAY);
		}

		if (test_type == TEST_TYPE_SPRINT_SENSOR) {
			register1 -= 1;
		}

setup_sprint_2:
		// Now let the user specify all the distances.
		for (i = 0; i < register1; ++i) {
			distance = 0;
			display_4bit_clear();
	
			display_4bit_set_position(0, 0);
			display_4bit_write_string("DISTANCE");
			display_4bit_set_position(9, 0);
			display_4bit_write_integer(i + 1, 2);
			display_4bit_write_string("/");
			display_4bit_write_integer(register1, 2);
			display_4bit_set_position(0, 1);
			display_4bit_write_string("IN METERS:");
			display_4bit_set_position(11, 1);
			display_4bit_write_integer(distance, 3);
		
			input_position = 0;
			wait_for_keypad_release();
			while ((key = keypad_scan()) != KEY_OK && key != KEY_NEXT) {
				wait_for_keypad_release();
				value = 10; // This is an invalid number.
		
				switch (key)
				{
				case KEY_SUB_INC:
					if (distance < 999) { ++distance; }
					input_position = 0;
					display_4bit_set_position(11, 1);
					display_4bit_write_integer(distance, 3);
					break;
					
				case KEY_POP_DEC:
					if (distance > register2) { --distance; }
					input_position = 0;
					display_4bit_set_position(11, 1);
					display_4bit_write_integer(distance, 3);
					break;
			
				// This menu entry causes the config to abort and returns us to
				// to the appropiate menu entry from whence we came.
				case KEY_PREV:
					//menu_state = (menu_states_t) register0;
					//return 0;
					//++register1; //TODO: commented out, why is it increased?
					reuse = 0;
					goto setup_sprint_1;
					break;
			
				// Abort alltogether and return to initial state.
				case KEY_CANCEL:
					menu_state = MENU_STATE_SPRINT;
					return 0;
					break;
		
				case KEY_0:
					value = 0;
					break;
		
				case KEY_1:
					value = 1;
					break;
		
				case KEY_2:
					value = 2;
					break;
		
				case KEY_3:
					value = 3;
					break;
		
				case KEY_4:
					value = 4;
					break;
		
				case KEY_5:
					value = 5;
					break;
		
				case KEY_6:
					value = 6;
					break;
		
				case KEY_7:
					value = 7;
					break;
		
				case KEY_8:
					value = 8;
					break;
		
				case KEY_9:
					value = 9;
					break;
			
				default:
					break;
				};
		
				// If value have been modified someone has input a digit. Process it.
				if (value != 10) {
					if (counter_ms > 2000) { input_position = 0; distance = 0; }
					// We must zero any value if this is the start of the keyboard
					// input sequence.
					if (input_position == 0) { distance = 0; }
					distance *= 10;
					distance += value;
					if (distance < DISTANCE_MIN) { distance = DISTANCE_MIN; }
					else if (distance > DISTANCE_MAX) { distance = DISTANCE_MAX; }
					
					display_4bit_set_position(11, 1);
					display_4bit_write_integer(distance, 3);
		
					// We only support three digits, so reset 'input_position' if needed.
					++input_position;
					if (input_position > 2) { input_position = 0; }

					counter_ms = 0;
				}
		
				delay_ms(KEY_PRESS_DELAY);
			}
		
			distances[i] = distance;
		}

		// Allow the user to input the player id into register3.
		register3 = 0;
	}

	display_4bit_clear();

	display_4bit_set_position(0, 0);
	display_4bit_write_string("SELECT PLAYER");
	display_4bit_set_position(0, 1);
	display_4bit_write_string("PLAYER:");
	display_4bit_set_position(8, 1);
	display_4bit_write_integer(register3, 2);

	input_position = 0;
	wait_for_keypad_release();
	while ((key = keypad_scan()) != KEY_OK && key != KEY_NEXT) {
		wait_for_keypad_release();	
		value = 10; // This is an invalid number.

		switch (key)
		{
		case KEY_SUB_INC:
			if (register3 < PLAYER_ID_MAX) { ++register3; }
			input_position = 0;
			display_4bit_set_position(8, 1);
			display_4bit_write_integer(register3, 2);
			break;
			
		case KEY_POP_DEC:
			if (register3 > PLAYER_ID_MIN) { --register3; }
			input_position = 0;
			display_4bit_set_position(8, 1);
			display_4bit_write_integer(register3, 2);
			break;
	
		// This menu entry causes the config to abort and returns us to
		// to the appropiate menu entry from whence we came.
		case KEY_PREV:
			goto setup_sprint_2;
			break;
	
		// Abort alltogether and return to initial state.
		case KEY_CANCEL:
			menu_state = MENU_STATE_SPRINT;
			return 0;
			break;

		case KEY_0:
			value = 0;
			break;

		case KEY_1:
			value = 1;
			break;

		case KEY_2:
			value = 2;
			break;

		case KEY_3:
			value = 3;
			break;

		case KEY_4:
			value = 4;
			break;

		case KEY_5:
			value = 5;
			break;

		case KEY_6:
			value = 6;
			break;

		case KEY_7:
			value = 7;
			break;

		case KEY_8:
			value = 8;
			break;

		case KEY_9:
			value = 9;
			break;
	
		default:
			break;
		};

		// If value have been modified someone has input a digit. Process it.
		if (value != 10) {
			if (counter_ms > 2000) { input_position = 0; register3 = 0; }
			// We must zero any value if this is the start of the keyboard
			// input sequence.
			if (input_position == 0) { register3 = 0; }
			register3 *= 10;
			register3 += value;
			if (register3 < PLAYER_ID_MIN) { register3 = PLAYER_ID_MIN; }
			else if (register3 > PLAYER_ID_MAX) { register3 = PLAYER_ID_MAX; }
			
			display_4bit_set_position(8, 1);
			display_4bit_write_integer(register3, 2);

			// We only support two digits, so reset 'input_position' if needed.
			++input_position;
			if (input_position > 1) { input_position = 0; }

			counter_ms = 0;
		}

		delay_ms(KEY_PRESS_DELAY);
	}

	return 1;
}

/*
Function: setup_jump_test

Description: This function allows the user to setup the parameters
    for a jump test.
Parameters:
    reuse: If not zero, only the player id will be modified.

    registers[0]: menu to return to for PREV press.
	registers[1]: menu to return to for CANCEL press. 
    registers[3]: player id

Return value:
    Nonzero if the test should proceed, zero if aborted.
*/
unsigned char setup_jump_test (unsigned char reuse)
{
	keys_t key;
	unsigned char value, i;

	// This variable is used to track where any direct number
	// entry happened in the input string.
	unsigned char input_position;

	register3 = 0;

	display_4bit_clear();

	display_4bit_set_position(0, 0);
	display_4bit_write_string("SELECT PLAYER");
	display_4bit_set_position(0, 1);
	display_4bit_write_string("PLAYER:");
	display_4bit_set_position(8, 1);
	display_4bit_write_integer(register3, 2);

	input_position = 0;
	wait_for_keypad_release();
	while ((key = keypad_scan()) != KEY_OK && key != KEY_NEXT) {
		wait_for_keypad_release();	
		value = 10; // This is an invalid number.

		switch (key)
		{
		case KEY_SUB_INC:
			if (register3 < PLAYER_ID_MAX) { ++register3; }
			input_position = 0;
			display_4bit_set_position(8, 1);
			display_4bit_write_integer(register3, 2);
			break;
			
		case KEY_POP_DEC:
			if (register3 > PLAYER_ID_MIN) { --register3; }
			input_position = 0;
			display_4bit_set_position(8, 1);
			display_4bit_write_integer(register3, 2);
			break;
	
		// This menu entry causes the config to abort and returns us to
		// to the appropiate menu entry from whence we came.
		case KEY_PREV:
			menu_state = (menu_states_t) register0;
			return 0;
			break;
	
		// Abort alltogether and return to initial state.
		case KEY_CANCEL:
			menu_state = (menu_states_t) register1;
			return 0;
			break;

		case KEY_0:
			value = 0;
			break;

		case KEY_1:
			value = 1;
			break;

		case KEY_2:
			value = 2;
			break;

		case KEY_3:
			value = 3;
			break;

		case KEY_4:
			value = 4;
			break;

		case KEY_5:
			value = 5;
			break;

		case KEY_6:
			value = 6;
			break;

		case KEY_7:
			value = 7;
			break;

		case KEY_8:
			value = 8;
			break;

		case KEY_9:
			value = 9;
			break;
	
		default:
			break;
		};

		// If value have been modified someone has input a digit. Process it.
		if (value != 10) {
			if (counter_ms > 2000) { input_position = 0; register3 = 0; }
			// We must zero any value if this is the start of the keyboard
			// input sequence.
			if (input_position == 0) { register3 = 0; }
			register3 *= 10;
			register3 += value;
			if (register3 < PLAYER_ID_MIN) { register3 = PLAYER_ID_MIN; }
			else if (register3 > PLAYER_ID_MAX) { register3 = PLAYER_ID_MAX; }
			
			display_4bit_set_position(8, 1);
			display_4bit_write_integer(register3, 2);

			// We only support two digits, so reset 'input_position' if needed.
			++input_position;
			if (input_position > 1) { input_position = 0; }

			counter_ms = 0;
		}

		delay_ms(KEY_PRESS_DELAY);
	}

	return 1;
}

/*
Function: wait_for_keypad_release

Description: Blocks execution until the keypad is released
    (until no keys are being held down).
*/
void wait_for_keypad_release (void)
{
	keys_t key, previous_key;

	previous_key = keypad_scan();
	while (((key = keypad_scan()) == previous_key) && (key != KEY_NONE)) {
		previous_key = key;
	}
}

/*
Function: prompt_for_save

Description: Displays the latest result and allows the user the
    option of saving the result to the EEPROM memory.

Return value: Nonzero is the user choose to save the result, else zero.

*/
unsigned char prompt_for_save (void)
{
	keys_t key;

	// Display result and ask user if he/she wants to save the result.
	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("RESULT:");
	realtime_update(results[0], 8, 0);
	display_4bit_set_position(0, 1);
	display_4bit_write_string("SAVE?");

	wait_for_keypad_release();

	for ( ; ; ) {
		key = keypad_scan();

		if (key == KEY_OK) {
			break;
		} else if (key == KEY_CANCEL) {
			return 0;
			break;
		}
	}

	display_4bit_set_position(0, 1);
	display_4bit_write_string("SAVING...");

	if (!record_store()) {
		display_4bit_set_position(0, 1);
		display_4bit_write_string("ERROR: MEM. FULL");
	} else {
		display_4bit_set_position(10, 1);
		display_4bit_write_string("OK");
	}

	delay_s(1);
	
	return 1;
}

#ifdef DEBUG
void dump_memory_to_serial (void)
{
	unsigned int location;
	unsigned char value;
	unsigned char buf[8];

	for (location = 0x0000; location < 64; ++location) {
		fm3164_read_memory(location, (unsigned char *) &value, sizeof(unsigned char));
		sprintf(buf, "0x%X ", value);
		uart_send_string(buf);
		//uart_send_character(value);
	}
}
#endif

void record_remove_player (unsigned char player_id)
{
	unsigned int end, size_of_record;
	unsigned int src, dst;
	unsigned char temp;

	current_record = 0x0002;

	// Read ot the value indicating the end of the memory.
	fm3164_read_memory(0x0000, (unsigned char *) &end, sizeof(unsigned int));
	
	while (current_record < end) {
		record_retrieve();

		if (global_record.player_id == player_id) {
			size_of_record = sizeof(record_t) + global_record.num_times * 
				(sizeof(unsigned long) + sizeof(unsigned int));
			dst = current_record - size_of_record;
			src = current_record;

			// Overwrite the record with the rest of the memory.
			while (src < end) {
				fm3164_read_memory_byte(src++, &temp);
				fm3164_write_byte_memory(dst++, temp);
			}

			end -= size_of_record;
			current_record -= size_of_record;
		}
	}

	fm3164_write_memory(0x0000, (unsigned char *) &end, sizeof(unsigned int));

}

/*
Parameters:
    record: Address of the record to modify.
    result_index: Index to the laptime / subresult to remove.
*/
void remove_result (unsigned int record, unsigned char result_index)
{
	unsigned int end;
	unsigned int src, dst;
	unsigned char temp;
	unsigned int results;

	// Read ot the value indicating the end of the memory.
	fm3164_read_memory(0x0000, (unsigned char *) &end, sizeof(unsigned int));

	results = record + sizeof(record_t);

	dst = results + (result_index * (sizeof(unsigned long) + sizeof(unsigned int)));
	src = results + ((result_index + 1) * (sizeof(unsigned long) + sizeof(unsigned int)));

	// Overwrite the record with the rest of the memory.
	while (src < end) {
		fm3164_read_memory_byte(src++, &temp);
		fm3164_write_byte_memory(dst++, temp);
	}

	// Decreaste the time count with 1 for the result stored in the EEPROM.
	fm3164_read_memory(record + sizeof(unsigned char) + sizeof(unsigned int) +
		sizeof(unsigned long) + sizeof(test_types_t), (unsigned char *) &temp,
		sizeof(unsigned char));

	temp -= 1;

	fm3164_write_memory(record + sizeof(unsigned char) + sizeof(unsigned int) +
		sizeof(unsigned long) + sizeof(test_types_t), (unsigned char *) &temp,
		sizeof(unsigned char));

	end -= (sizeof(unsigned long) + sizeof(unsigned int));
	fm3164_write_memory(0x0000, (unsigned char *) &end, sizeof(unsigned int));
}

unsigned char test_endurance_beep (void)
{
	unsigned long stage;
	unsigned char level, completed, laps;
	unsigned char buffer[8];
	unsigned long counter_ms2;

	unsigned char i, num_results, value, input_position, player_id, wait_for_more_digits;
	unsigned long key_timer;
	unsigned char players[PLAYER_COUNT];
	unsigned char levels[PLAYER_COUNT];
	unsigned char stages[PLAYER_COUNT];

	keys_t key, old_key;

	register0 = MENU_STATE_ENDURANCE_BEEP;
	laps = 0;

	key_timer = 0;
	input_position = 0;
	wait_for_more_digits = 0;
	old_key = KEY_NONE;

	level = 0;
	stage = 0;
	key = KEY_NONE;

	wait_for_keypad_release();

	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("PRESS OK");
	display_4bit_set_position(0, 1);
	display_4bit_write_string("TO START");

	// Wait for user to proceed ('ok'), abort ('cancel') or go back ('prev').
	while ((key = keypad_scan()) != KEY_OK) {
		if (key == KEY_CANCEL) { menu_state = MENU_STATE_ENDURANCE; return 0; }
		else if (key == KEY_PREV) { return 0; }
	}

	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("STG: ");
	uitoa_dec(buffer, (unsigned int) stage + 1, 2);
	buffer[2] = ' ';
	buffer[3] = '\0';
	display_4bit_write_string(buffer);
		
	display_4bit_write_string("LVL: ");
	uitoa_dec(buffer, (unsigned int) level + 1, 2);	
	buffer[2] = '\0';
	display_4bit_write_string(buffer);

	display_4bit_set_position(0, 1);
	display_4bit_write_string("LAPS: ");
	uitoa_dec(buffer, laps +1 , 3);
	buffer[3] = '\0';
	display_4bit_write_string(buffer);

	wait_for_keypad_release();

	// Store the current time.
	fm3164_get_date(&global_record.date);

	// Emit a starting beep and reset the counter.
	portc.RC2 = 1;
	counter_ms = 0;
	delay_ms(START_BEEP_DELAY);
	portc.RC2 = 0;

	i = 0;
	for (stage = 0; stage < 20; ++stage) {
		counter_ms2 = 0;
		level = 0;

		// Redraw screen at stage change.
		display_4bit_set_position(0, 0);
		display_4bit_write_string("STG: ");
		uitoa_dec(buffer, stage + 1, 2);
		buffer[2] = ' ';
		buffer[3] = '\0';
		display_4bit_write_string(buffer);
		
		display_4bit_write_string("LVL: ");
		uitoa_dec(buffer, (unsigned int) level + 1, 2);
		buffer[2] = '\0';
		display_4bit_write_string(buffer);

		display_4bit_set_position(0, 1);
		display_4bit_write_string("LAPS: ");
		uitoa_dec(buffer, laps + 1, 3);
		buffer[3] = '\0';
		display_4bit_write_string(buffer);

		while (counter_ms2 < 60000) {
			if (counter_ms > (8471 - (stage * 229))) {
				++level;
				++laps;
				portc.RC2 = 1;
				delay_ms(NOTICE_BEEP_DELAY);
				portc.RC2 = 0;
				
				counter_ms2 += counter_ms;
				counter_ms = 0;

				display_4bit_set_position(0, 0);
				display_4bit_write_string("STG: ");
				uitoa_dec(buffer, stage + 1, 2);
				buffer[2] = ' ';
				buffer[3] = '\0';
				display_4bit_write_string(buffer);
				
				display_4bit_write_string("LVL: ");
				uitoa_dec(buffer, (unsigned int) level + 1, 2);
				buffer[2] = '\0';
				display_4bit_write_string(buffer);

				display_4bit_set_position(0, 1);
				display_4bit_write_string("LAPS: ");
				uitoa_dec(buffer, laps + 1, 3);
				buffer[3] = '\0';
				display_4bit_write_string(buffer);
			}

			if ((key = keypad_scan()) == KEY_CANCEL) {
				menu_state = MENU_STATE_ENDURANCE;
				return 0;
			} else if (key == KEY_PREV) {
				return 0;
			} else if (key == KEY_OK) {
				stage = 20; // Hack, breaks out of the while() loop.
				break;
			}

			value = 10; // This is an invalid number.

			if (key != old_key) {
				old_key = key;
				delay_ms(200);
			switch (key)
			{
			case KEY_0:
				value = 0;
				break;
	
			case KEY_1:
				value = 1;
				break;
	
			case KEY_2:
				value = 2;
				break;
	
			case KEY_3:
				value = 3;
				break;
	
			case KEY_4:
				value = 4;
				break;
	
			case KEY_5:
				value = 5;
				break;
	
			case KEY_6:
				value = 6;
				break;
	
			case KEY_7:
				value = 7;
				break;
	
			case KEY_8:
				value = 8;
				break;
	
			case KEY_9:
				value = 9;
				break;
		
			default:
				break;
			};
	
			// If value have been modified someone has input a digit. Process it.
			if (value != 10) {
				// New value when 'input_position' == 0 means a new value is being input.
				// Clear the old value.
				if (input_position == 0) { player_id = 0; }
	
				player_id *= 10;
				player_id += value;
				
				// Clamp input values to legal ranges.
				if (player_id < PLAYER_ID_MIN) { player_id = PLAYER_ID_MIN; }
				else if (player_id > PLAYER_ID_MAX) { player_id = PLAYER_ID_MAX; }
				
				display_4bit_set_position(14, 1);
				display_4bit_write_integer(player_id, 2);
	
				// We only support two digits, so reset 'input_position' if needed.
				++input_position;
				if (input_position > 1) { input_position = 0; }

				wait_for_more_digits = 1;

			}
			}

			// If the key timer expires, take input and store it.
			if (key_timer > 1500) {
				
				if (i < PLAYER_COUNT) {
					players[i] = player_id;
					levels[i] = level;
					stages[i] = stage;
				}

				#ifdef DEBUG_ENDURANCE_BEEP
				uart_send_string("Key input, id: ");
				uitoa_dec(buffer, player_id, 2);
				uart_send_string(buffer);
				uart_send_string(", level: ");
				uitoa_dec(buffer, level, 2);
				uart_send_string(buffer);
				uart_send_string(", stage: ");
				uitoa_dec(buffer, stage, 2);
				uart_send_string(buffer);
				uart_send_string("\r\n");
				#endif

				input_position = 0; player_id = 0; key_timer = 0;
				++i;

				display_4bit_set_position(14, 1);
				display_4bit_write_string("--");



				wait_for_more_digits = 0;
			}
			
			if (wait_for_more_digits) { ++key_timer; }
		}

		// This is part of the double-beeping.
		delay_ms(NOTICE_BEEP_DELAY);
		portc.RC2 = 1;
		delay_ms(NOTICE_BEEP_DELAY);
		portc.RC2 = 0;
		
	}

	num_results = i;

	wait_for_keypad_release();

	if (num_results > 0) {
		display_4bit_clear();
		display_4bit_set_position(0, 0);
		display_4bit_write_string("SAVE RESULTS?");

		while ((key = keypad_scan()) != KEY_OK) {
			if (key == KEY_CANCEL) { return 0; }
			else if (key == KEY_PREV) { return 0; }
		}
	// If no results were accumulated, just go back to the main menu entry.
	} else { return 0; }

	// If the user pressed OK then store the result times to EEPROM.
	#ifdef DEBUG_ENDURANCE_BEEP
	uart_send_string("Saving to EEPROM, number of records: ");
	uitoa_dec(buffer, num_results, 2);
	uart_send_string(buffer);
	uart_send_string("\r\n");
	#endif

	for (i = 0; i < num_results; ++i) {
		global_record.player_id = players[i];
		results[0] = stages[i] + 1;
		distances[0] = levels[i] + 1;
		global_record.type = TEST_TYPE_ENDURANCE_BEEP;
		global_record.num_times = 1;

		#ifdef DEBUG_ENDURANCE_BEEP
		uart_send_string("#");
		uitoa_dec(buffer, i, 2);
		uart_send_string(buffer);
		uart_send_string(" player id: ");
		uitoa_dec(buffer, global_record.player_id, 2);
		uart_send_string(buffer);
		uart_send_string(", level: ");
		uitoa_dec(buffer, (unsigned int) distances[0], 2);
		uart_send_string(buffer);
		uart_send_string(", stage: ");
		uitoa_dec(buffer, results[0], 2);
		uart_send_string(buffer);
		uart_send_string("\r\n");
		#endif

		display_4bit_set_position(0, 1);
		display_4bit_write_string("SAVING...");

		if (!record_store()) {
			display_4bit_set_position(0, 1);
			display_4bit_write_string("ERROR: MEM. FULL");
		} else {
			display_4bit_set_position(10, 1);
			display_4bit_write_string("OK");
		}
	}
	
	// Returning 0 means another test should not automatically performed.
	return 0;
}

/*
Function: test_endurance_multi

*/
unsigned char test_endurance_multi (void)
{
	unsigned char i, j, best_result, terminate;
	unsigned long temp;
	keys_t key;
	unsigned char buffer[8];
	unsigned long run_times[MAX_SENSORS];
	unsigned int rest_times[MAX_SENSORS];

	#ifdef DEBUG_ENDURANCE_MULTI
	unsigned char buffer2[16];
	#endif

	menu_state = MENU_STATE_ENDURANCE_MULTI;
	best_result = 0;
	terminate = 0;

	wait_for_keypad_release();

	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("RUN! PRESS OK");
	display_4bit_set_position(0, 1);
	display_4bit_write_string("WHEN FINISHED");

	// Store the current time.
	fm3164_get_date(&global_record.date);

	// Wait for the user trigger the first sensor.
	while (portb.RB0) {
		key = keypad_scan();
		if (key == KEY_CANCEL) { return 1; }
		else if (key == KEY_PREV) { return 2; }
	}

	portc.RC2 = 1;
	counter_ms = 0;
	delay_ms(START_BEEP_DELAY);
	portc.RC2 = 0;

	display_4bit_clear();

	// We can only use MAX_SENSORS - 1 since one slot is lost to the dummy value.
	for (i = 0; i < MAX_SENSORS - 1 && (!terminate); ++i) {
		display_4bit_set_position(0, 0);
		display_4bit_write_string("SPRINT: ");
		uitoa_dec(buffer, i + 1, 2);
		buffer[2] = '\0';
		display_4bit_write_string(buffer);

		display_4bit_set_position(0, 1);
		display_4bit_write_string("RUN: ");
		
		// Debouncing, require a low level before waiting for next signal.
		while (!portb.RB0);

		// Wait for sensor signal.
		while (portb.RB0 && !terminate) {
			key = keypad_scan();
			if (key == KEY_OK) { terminate = 1; }
			else if (key == KEY_CANCEL) { return 1; }
			else if (key == KEY_PREV) { return 2; }
			realtime_update(counter_ms, 6, 1);
		}

		if (!terminate) {
			portc.RC2 = 1;

			run_times[i] = counter_ms;
			counter_ms = 0;
			delay_ms(NOTICE_BEEP_DELAY);
			portc.RC2 = 0;
		} else {
			// If 'termination' occured during a run phase we won't have
			// a valid time to write to memory, thus we decrement i.
			--i;
		}

		display_4bit_set_position(0, 1);
		display_4bit_write_string("REST:");

		// Debouncing, require a low level before waiting for next signal.
		while (!portb.RB0 && !terminate);

		// Wait for user to rest up and trigger sensor again.
		while (portb.RB0 && !terminate) {
			key = keypad_scan();
			if (key == KEY_OK) { terminate = 1; }
			else if (key == KEY_CANCEL) { return 1; }
			else if (key == KEY_PREV) { return 2; }
			realtime_update(counter_ms, 6, 1);
		}

		portc.RC2 = 1;
		rest_times[i] = (unsigned int) (counter_ms / 1000);
		counter_ms = 0;
		delay_ms(NOTICE_BEEP_DELAY);
		portc.RC2 = 0;

	}

	// Sound the alarm to indicate last measuring.
	portc.RC2 = 1;
	delay_ms(NOTICE_BEEP_DELAY);
	portc.RC2 = 0;

	global_record.num_times = i + 1; // Add one since we have a dummy.
	global_record.type = TEST_TYPE_ENDURANCE_MULTI;
	global_record.player_id = register3;

	// First entry in the 'distances' array is the test distance.
	// The value below ('results[0]') is a dummy value.
	// The number of results recorded is stored in 'i'.
	// We must increase global_record.num_times with 1 in order to accomodate the extra data.
	results[0] = 0;
	for (j = 0; j < i; ++j) {
		distances[j + 1] = rest_times[j];
		results[j + 1] = run_times[j];

		#ifdef DEBUG_ENDURANCE_MULTI
		sprintf(buffer2, "runtime %u: ", (j + 1));
		uart_send_string(buffer2);
		sprintf32(buffer2, "%u", run_times[j]);
		uart_send_string(buffer2);
		uart_send_string("\r\n");
	
		sprintf(buffer2, "resttime %u: ", (j + 1));
		uart_send_string(buffer2);
		sprintf(buffer2, "%u", rest_times[j]);
		uart_send_string(buffer2);
		uart_send_string("\r\n");
		#endif

		if (run_times[j] < run_times[best_result]) { best_result = j; }
	}

	// We only show the result screen if we have more than one result.
	if (i > 0) {
		// Display result and ask user if he/she wants to save the result.
		display_4bit_clear();
		display_4bit_set_position(0, 0);
		display_4bit_write_string("RESULT: ");
		time_long_to_string(buffer, 10, (unsigned long) run_times[best_result]);
		display_4bit_write_string(buffer);
		display_4bit_write_string(" s");
		display_4bit_set_position(0, 1);
		display_4bit_write_string("SAVE?");
	
		display_4bit_set_position(10, 1);
		uitoa_dec(buffer, best_result + 1, 2);
		buffer[2] = '/';
		uitoa_dec(buffer + 3, j, 2);
		buffer[5] = '\0';
		display_4bit_write_string(buffer);
	
		wait_for_keypad_release();
	
		for ( ; ; ) {
			key = keypad_scan();
	
			if (key == KEY_OK) {
				break;
			} else if (key == KEY_CANCEL) {
				return 1;
				break;
			}
		}
	
		display_4bit_clear();
		display_4bit_set_position(0, 1);
		display_4bit_write_string("SAVING...");
	
		if (!record_store()) {
			display_4bit_set_position(0, 1);
			display_4bit_write_string("ERROR: MEM. FULL");
		} else {
			display_4bit_set_position(10, 1);
			display_4bit_write_string("OK");
		}
	
	// If no results were recorded, insult user...
	} else {
		display_4bit_clear();
		display_4bit_set_position(0, 0);
		display_4bit_write_string("NO RESULTS");
	}

	delay_s(2);
	
	return 1;
}

/*
Function: setup_endurance_test

register0: Menu to return to when PREV is pressed.
register1: Menu to return to when CANCEL is pressed.
register2: Distance
register3: Player ID.

*/
unsigned char setup_endurance_test (unsigned char reuse)
{
	keys_t key;
	unsigned char value, i;
	unsigned int distance;

	#ifdef DEBUG_ENDURANCE_SETUP
	char buffer[16];
	#endif

	// This variable is used to track where any direct number
	// entry happened in the input string.
	unsigned char input_position;

	// If reuse is 1, then skip entering a distance.
	if (!reuse) {
setup_endurance_1:
		// Now let the user specify all the distances.
		for (i = 0; i < 1; ++i) {
			distance = 0;
			display_4bit_clear();
	
			display_4bit_set_position(0, 0);
			display_4bit_write_string("DISTANCE");
			display_4bit_set_position(0, 1);
			display_4bit_write_string("IN METERS:");
			display_4bit_set_position(11, 1);
			display_4bit_write_integer(distance, 3);
		
			input_position = 0;
			wait_for_keypad_release();
			counter_ms = 0;
			while ((key = keypad_scan()) != KEY_OK && key != KEY_NEXT) {
				wait_for_keypad_release();
				value = 10; // This is an invalid number.
		
				switch (key)
				{
				case KEY_SUB_INC:
					if (distance < 999) { ++distance; }
					input_position = 0;
					display_4bit_set_position(11, 1);
					display_4bit_write_integer(distance, 3);
					break;
					
				case KEY_POP_DEC:
					if (distance > register2) { --distance; }
					input_position = 0;
					display_4bit_set_position(11, 1);
					display_4bit_write_integer(distance, 3);
					break;
			
				// This menu entry causes the config to abort and returns us to
				// to the appropiate menu entry from whence we came.
				case KEY_PREV:
					#ifdef DEBUG_ENDURANCE_SETUP
					uart_send_string("aborting from PREV press\r\n");
					sprintf(buffer, "r0: %u", register0);
					uart_send_string(buffer);
					sprintf(buffer, "r1: %u", register1);
					uart_send_string(buffer);
					sprintf(buffer, "r2: %u", register2);
					uart_send_string(buffer);
					sprintf(buffer, "r3: %u", register3);
					uart_send_string(buffer);
					uart_send_string("\r\n");
					#endif

					menu_state = (menu_states_t) register0;
					return 0;
					break;
			
				// Abort alltogether and return to initial state.
				case KEY_CANCEL:
					menu_state = (menu_states_t) register1;
					return 0;
					break;
		
				case KEY_0:
					value = 0;
					break;
		
				case KEY_1:
					value = 1;
					break;
		
				case KEY_2:
					value = 2;
					break;
		
				case KEY_3:
					value = 3;
					break;
		
				case KEY_4:
					value = 4;
					break;
		
				case KEY_5:
					value = 5;
					break;
		
				case KEY_6:
					value = 6;
					break;
		
				case KEY_7:
					value = 7;
					break;
		
				case KEY_8:
					value = 8;
					break;
		
				case KEY_9:
					value = 9;
					break;
			
				default:
					break;
				};
		
				// If value have been modified someone has input a digit. Process it.
				if (value != 10) {
					if (counter_ms > 2000) { input_position = 0; distance = 0; }
					// We must zero any value if this is the start of the keyboard
					// input sequence.
					if (input_position == 0) { distance = 0; }
					distance *= 10;
					distance += value;
					if (distance < DISTANCE_MIN) { distance = DISTANCE_MIN; }
					else if (distance > DISTANCE_MAX) { distance = DISTANCE_MAX; }
					
					display_4bit_set_position(11, 1);
					display_4bit_write_integer(distance, 3);
		
					// We only support three digits, so reset 'input_position' if needed.
					++input_position;
					if (input_position > 2) { input_position = 0; }

					counter_ms = 0;
				}
		
				delay_ms(KEY_PRESS_DELAY);
			}
		
			distances[i] = distance;
		}
	
	}

	// Allow the user to input the player id into register3.
	register3 = 0; // CHECKME: zero here or outside?

	display_4bit_clear();

	display_4bit_set_position(0, 0);
	display_4bit_write_string("SELECT PLAYER");
	display_4bit_set_position(0, 1);
	display_4bit_write_string("PLAYER:");
	display_4bit_set_position(8, 1);
	display_4bit_write_integer(register3, 2);

	input_position = 0;
	wait_for_keypad_release();
	counter_ms = 0;
	while ((key = keypad_scan()) != KEY_OK && key != KEY_NEXT) {
		wait_for_keypad_release();	
		value = 10; // This is an invalid number.

		switch (key)
		{
		case KEY_SUB_INC:
			if (register3 < PLAYER_ID_MAX) { ++register3; }
			input_position = 0;
			display_4bit_set_position(8, 1);
			display_4bit_write_integer(register3, 2);
			break;
			
		case KEY_POP_DEC:
			if (register3 > PLAYER_ID_MIN) { --register3; }
			input_position = 0;
			display_4bit_set_position(8, 1);
			display_4bit_write_integer(register3, 2);
			break;
	
		// This menu entry causes the config to abort and returns us to
		// to the appropiate menu entry from whence we came.
		case KEY_PREV:
			goto setup_endurance_1;
			break;
	
		// Abort alltogether and return to initial state.
		case KEY_CANCEL:
			menu_state = (menu_states_t) register1;
			return 0;
			break;

		case KEY_0:
			value = 0;
			break;

		case KEY_1:
			value = 1;
			break;

		case KEY_2:
			value = 2;
			break;

		case KEY_3:
			value = 3;
			break;

		case KEY_4:
			value = 4;
			break;

		case KEY_5:
			value = 5;
			break;

		case KEY_6:
			value = 6;
			break;

		case KEY_7:
			value = 7;
			break;

		case KEY_8:
			value = 8;
			break;

		case KEY_9:
			value = 9;
			break;
	
		default:
			break;
		};

		// If value have been modified someone has input a digit. Process it.
		if (value != 10) {
			if (counter_ms > 2000) { input_position = 0; register3 = 0; }
			// We must zero any value if this is the start of the keyboard
			// input sequence.
			if (input_position == 0) { register3 = 0; }
			register3 *= 10;
			register3 += value;
			if (register3 < PLAYER_ID_MIN) { register3 = PLAYER_ID_MIN; }
			else if (register3 > PLAYER_ID_MAX) { register3 = PLAYER_ID_MAX; }
			
			display_4bit_set_position(8, 1);
			display_4bit_write_integer(register3, 2);

			// We only support two digits, so reset 'input_position' if needed.
			++input_position;
			if (input_position > 1) { input_position = 0; }

			counter_ms = 0;
		}

		delay_ms(KEY_PRESS_DELAY);
	}

	return 1;
}

/*
Function: menu_clear_memory

Registers:
    register0: Menu to return to after clear.
*/
void menu_clear_memory (void)
{
	keys_t key;
	
	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("CLEAR MEMORY?");

	wait_for_keypad_release();
	
	while (1) {
		key = keypad_scan();

		if (key == KEY_CANCEL) {
			menu_state = MENU_STATE_SETUP_MEMORY;
			break;

		} else if (key == KEY_OK) {
			init_eeprom();
			display_4bit_clear();
			display_4bit_set_position(0, 0);
			display_4bit_write_string("MEMORY CLEARED");
			delay_s(1);
			break;
		}
	}

	wait_for_keypad_release();
}

/*
	register0: Menu to return to for PREV press.
	register1: Menu to return to for CANCEL press.
	register3: Player ID to remove.
*/
unsigned char menu_clear_player (void)
{
	keys_t key;
	char buffer[8];

	uitoa_dec(buffer, register3, 2);
	buffer[2] = '\0';
	
	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("CLEAR PLAYER ");
	display_4bit_write_string(buffer);
	display_4bit_write_string("?");

	wait_for_keypad_release();
	
	while (1) {
		key = keypad_scan();

		// If CANCEL is pressed, abort completely and return to menu.
		if (key == KEY_CANCEL) {
			//menu_state = (menu_states_t) register1;
			return 1;

		// If PREV is pressed, return to the setup function to select another player.
		} else if (key == KEY_PREV) {
			return 1;

		} else if (key == KEY_OK) {
			record_remove_player(register3);
			display_4bit_clear();
			display_4bit_set_position(0, 0);
			display_4bit_write_string("PLAYER ");
			display_4bit_write_string(buffer);
			display_4bit_set_position(0, 1);
			display_4bit_write_string("REMOVED");
			delay_s(1);
			break;
		}
	}

	wait_for_keypad_release();

	return 1;
}

void menu_memory_status (void)
{
	keys_t key;
	unsigned char buffer[16];
	unsigned int end, free;

	// Read ot the value indicating the end of the memory.
	fm3164_read_memory(0x0000, (unsigned char *) &end, sizeof(unsigned int));

	end -= 2;
	free = FM3164_MEMORY_SIZE - end;

	sprintf(buffer, "USED: %u", end);

	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string(buffer);

	sprintf(buffer, "FREE: %u", free);

	display_4bit_set_position(0, 1);
	display_4bit_write_string(buffer);

	wait_for_keypad_release();

	while (1) {
		key = keypad_scan();
		if (key == KEY_CANCEL) {
			menu_state = MENU_STATE_SETUP;
			break;
		} else if (key == KEY_PREV) {
			menu_state = MENU_STATE_SETUP_MEMORY_STATUS;
			break;
		}
	}

	wait_for_keypad_release();
}

void menu_setup_date (void)
{
	keys_t key;
	unsigned char position;
	fm3164_date_t date;
	unsigned char temp, redraw;
	unsigned char values[5]; // Vector with elements for year, month, day, hours & minutes.
	unsigned char blink_timer, blink_toggle;
	
	#ifdef MENU_DEBUG
	unsigned char buffer[16];
	#endif

	position = 0;
	blink_timer = 5;
	blink_toggle = 0;

	wait_for_keypad_release();

	fm3164_get_date(&date);
	
	// Begin by clearing display, load existing values into vector
	// and draw the vector values to screen.
	display_4bit_clear();
	display_4bit_set_position(0, 0);
	//display_4bit_write_string("DATE: 20");

	values[0] = date.year;
	//display_4bit_write_integer(values[0], 2);

	#ifdef MENU_DEBUG
	sprintf(buffer, "values[0]: %u\r\n", values[0]);
	uart_send_string(buffer);
	#endif

	values[1] = date.month;
	//display_4bit_write_integer(values[1], 2);

	#ifdef MENU_DEBUG
	sprintf(buffer, "values[1]: %u\r\n", values[1]);
	uart_send_string(buffer);
	#endif

	values[2] = date.day;
	//display_4bit_write_integer(values[2], 2);

	#ifdef MENU_DEBUG
	sprintf(buffer, "values[2]: %u\r\n", values[2]);
	uart_send_string(buffer);
	#endif

	//display_4bit_set_position(0, 1);
	//display_4bit_write_string("TIME: ");

	values[3] = date.hours;
	//display_4bit_write_integer(values[3], 2);

	#ifdef MENU_DEBUG
	sprintf(buffer, "values[3]: %u\r\n", values[3]);
	uart_send_string(buffer);
	#endif

	//display_4bit_write_character(':');

	values[4] = date.minutes;
	//display_4bit_write_integer(values[4], 2);

	#ifdef MENU_DEBUG
	sprintf(buffer, "values[4]: %u\r\n", values[4]);
	uart_send_string(buffer);
	#endif

	while (1) {
		redraw = 0;
		
		if (blink_timer++ > 2) {
			redraw = 1;
			blink_toggle = !blink_toggle;
			blink_timer = 0;
		}

		key = keypad_scan();
		if (key == KEY_CANCEL) {
			menu_state = MENU_STATE_SETUP;
			break;

		} else if (key == KEY_NEXT) {
			if (position < 9) { ++position; }
			blink_timer = 5;
			blink_toggle = 0;

		} else if (key == KEY_PREV) {
			if (position > 0) { --position; }
			blink_timer = 5;
			blink_toggle = 0;

		} else if (key == KEY_SUB_INC) {

			redraw = 1;
			
			// For odd numbers, add one.
			if (position % 2) {
				values[ (unsigned char) (position / 2) ] += 1;
			// For even numbers, add ten.
			} else {
				values[ (unsigned char) (position / 2) ] += 10;
			}

		} else if (key == KEY_POP_DEC) {

			redraw = 1;

			// For odd numbers, subtract one.
			if (position % 2) {
				values[ (unsigned char) (position / 2)] -= 1;
			// For even numbers, subtract ten.
			} else {
				values[ (unsigned char) (position / 2) ] -= 10;
			}

		} else if (key == KEY_OK) {
			date.year = values[0];
			date.month = values[1];
			date.day = values[2];
			date.hours = values[3];
			date.minutes = values[4];
			date.seconds = 0;
			fm3164_set_date(&date);

			menu_state = MENU_STATE_SETUP_DATE;
			break;
		}

		// Check values and clamp to valid ones.
		// Years (last two digits)
		if (values[0] > 99) { values[0] = 99; }

		// Months
		if (values[1] > 12) { values[1] = 12; }
		if (values[1] < 1) { values[1] = 1; }

		// Days
		if (values[2] > 31) { values[2] = 31; }
		if (values[2] < 1) { values[2] = 1; }
		
		// Hours and minutes
		if (values[3] > 23) { values[3] = 23; }
		if (values[4] > 59) { values[4] = 59; }

		if (redraw) {
			display_4bit_clear();
			display_4bit_set_position(0, 0);
			display_4bit_write_string("DATE: 20");
		
			display_4bit_write_integer(values[0], 2);
			display_4bit_write_character('-');
			display_4bit_write_integer(values[1], 2);
			display_4bit_write_character('-');
			display_4bit_write_integer(values[2], 2);
		
			display_4bit_set_position(0, 1);
			display_4bit_write_string("TIME: ");
		
			display_4bit_write_integer(values[3], 2);
		
			display_4bit_write_character(':');
		
			display_4bit_write_integer(values[4], 2);
		
			if (blink_toggle) {
				switch (position)
				{
				case 0:
					display_4bit_set_position(8, 0);
					break;
				case 1:
					display_4bit_set_position(9, 0);
					break;
				case 2:
					display_4bit_set_position(11, 0);
					break;
				case 3:
					display_4bit_set_position(12, 0);
					break;
				case 4:
					display_4bit_set_position(14, 0);
					break;
				case 5:
					display_4bit_set_position(15, 0);
					break;
				case 6:
					display_4bit_set_position(6, 1);
					break;
				case 7:
					display_4bit_set_position(7, 1);
					break;
				case 8:
					display_4bit_set_position(9, 1);
					break;
				case 9:
					display_4bit_set_position(10, 1);
					break;
				}
				
				display_4bit_write_character(0xff);
			}

			redraw = 0;
		}

		delay_ms(200);
	}

}

void menu_results_sprint (void)
{
	register0 = MENU_STATE_RESULTS_SPRINT;
	register1 = MENU_STATE_RESULTS;
	register3 = 0;

	current_record = 0x0002;
	if (setup_jump_test(0)) {
		while (display_time_results(TEST_TYPE_SPRINT_SENSOR | TEST_TYPE_SPRINT_RANDOM
		| TEST_TYPE_SPRINT_SIGNAL | TEST_TYPE_SPRINT_SWITCH, register3)) {
			if (!setup_jump_test(1)) { break; } }
	}

}

void menu_results_jump (void)
{
	register0 = MENU_STATE_RESULTS_JUMP;
	register1 = MENU_STATE_RESULTS;
	register3 = 0;

	current_record = 0x0002;
	if (setup_jump_test(0)) {
		while (display_time_results(TEST_TYPE_JUMP_SINGLE | TEST_TYPE_JUMP_MULTIPLE, register3)) {
			if (!setup_jump_test(1)) { break; } }
	}
}

void menu_results_endurance (void)
{
	register0 = MENU_STATE_RESULTS_ENDURANCE;
	register1 = MENU_STATE_RESULTS;
	register3 = 0;

	current_record = 0x0002;
	if (setup_jump_test(0)) {
		while (display_time_results(TEST_TYPE_ENDURANCE_BEEP| TEST_TYPE_ENDURANCE_MULTI, register3)) {
			if (!setup_jump_test(1)) { break; } }
	}
}

/*
Function: input_player

*/
keys_t input_player (unsigned char *player_id)
{
	unsigned int player;
	keys_t key;

	player = 0;

	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("SELECT PLAYER");
	
	display_4bit_set_position(0, 1);
	display_4bit_write_string("PLAYER");
	display_4bit_set_position(7, 1);
	display_4bit_write_integer(player, 2);

	wait_for_keypad_release();

	while (1) {
		key = keypad_scan();
		
		// 'cancel' or 'prev' aborts and returns.
		if (key == KEY_CANCEL) {
			break;

		} else if (key == KEY_PREV) {
			break;

		// 'sub/inc' increases player id
		} else if (key == KEY_SUB_INC) {
			if (player < PLAYER_ID_MAX) {
				++player;
				display_4bit_set_position(7, 1);
				display_4bit_write_integer(player, 3);
			}

		// 'pop/dec' decreases player id
		} else if (key == KEY_POP_DEC) {
			if (player > PLAYER_ID_MIN) {
				--player;
				display_4bit_set_position(7, 1);
				display_4bit_write_integer(player, 3);
			}

		// 'ok' stores the player id and returns
		} else if (key == KEY_OK) {
			*player_id = player;
			break;
		}

		delay_ms(200);
	}

	return key;
}

/*
Function: display_time_results
Description: Displays the current result and allows the user to
    view the different results.
*/
unsigned char display_time_results (test_types_t type, unsigned char player_id)
{
	keys_t key;
	unsigned char redraw, num_results, index, i;
	unsigned char buffer[32];
	unsigned int end;
	unsigned int best_result; // Index variable to remember best result.

	current_record = 0x0002;

	// Read ot the value indicating the end of the memory.
	fm3164_read_memory(0x0000, (unsigned char *) &end, sizeof(unsigned int));

	// Request caching of up to 'NUM_CACHED_RESULTS' in the 'result_adresses' array,
	// all of which must match 'type' and 'player_id'.
	num_results = cache_results(result_addresses, type, player_id);
	if (num_results == 0) {
		display_4bit_clear();
		display_4bit_set_position(0, 0);
		display_4bit_write_string("NO RESULTS");
		delay_s(2);
		return 1;
	}

	#ifdef DEBUG_RESULTS
	sprintf(buffer, "numrslts: %u\r\n", num_results);
	uart_send_string(buffer);
	sprintf(buffer, "end: 0x%X\r\n", end);
	uart_send_string(buffer);
	sprintf(buffer, "cur: 0x%X\r\n", current_record);
	uart_send_string(buffer);
	#endif

	redraw = 1;
	index = 0;

	wait_for_keypad_release();

	// Force refresh for first entry.
	current_record = result_addresses[index];
	record_retrieve();

	while (1) {
		key = keypad_scan();
		
		// 'cancel' requests deletion of the result
		if (key == KEY_CANCEL) {
			#ifdef DEBUG_RESULTS
			uart_send_string("cancel pressed\r\n");
			#endif
			return 1;

		} else if (key == KEY_PREV) {
			#ifdef DEBUG_RESULTS
			uart_send_string("prev pressed\r\n");
			#endif
			return 1;

		// 'sub/inc' increases the result index
		} else if (key == KEY_SUB_INC) {

			#ifdef DEBUG_RESULTS
			uart_send_string("sub/inc pressed\r\n");
			#endif

			if (index < num_results - 1) {
				++index;
				current_record = result_addresses[index];
				record_retrieve();
				redraw = 1;
			}

			#ifdef DEBUG_RESULTS
			sprintf(buffer, "exit search, end: 0x%X\r\n", end);
			uart_send_string(buffer);
			sprintf(buffer, "cur: 0x%X\r\n", current_record);
			uart_send_string(buffer);
			#endif

		// 'pop/dec' decreases the result index
		} else if (key == KEY_POP_DEC) {

			#ifdef DEBUG_RESULTS
			uart_send_string("pop/dec pressed\r\n");
			#endif

			if (index > 0) {
				--index;
				current_record = result_addresses[index];
				record_retrieve();
				redraw = 1;
			}

			#ifdef DEBUG_RESULTS
			sprintf(buffer, "exit search, end: 0x%X\r\n", end);
			uart_send_string(buffer);
			sprintf(buffer, "cur: 0x%X\r\n", current_record);
			uart_send_string(buffer);
			#endif

		// Delete the result!
		} else if (key == KEY_OK) {

			#ifdef DEBUG_RESULTS
			uart_send_string("ok pressed\r\n");
			#endif

			// Allow the user to remove the record (returns 1 if a record was removed).
			if (menu_record_remove(result_addresses[index])) {
				// Read ot the value indicating the end of the memory.
				fm3164_read_memory(0x0000, (unsigned char *) &end, sizeof(unsigned int));

				// Request caching of up to 'NUM_CACHED_RESULTS' in the 'result_adresses' array,
				// all of which must match 'type' and 'player_id'.
				num_results = cache_results(result_addresses, type, player_id);

				// If the user deletes the last record we clamp the index to the nearerst
				// valid value. If there are no results left we abort.
				if (index >= num_results && num_results > 0) {
					index = num_results - 1;
				} else if (num_results == 0) {
					display_4bit_clear();
					display_4bit_set_position(0, 0);
					display_4bit_write_string("NO RESULTS");
					delay_s(2);
					return 1;
				}
				redraw = 1;
			}
		}

		if (redraw) {

			#ifdef DEBUG_RESULTS
			uart_send_string("redrawed\r\n");
			#endif
		
			display_4bit_clear();
		
			// Display the test timestamp at the upper row
			buffer[0] = '2'; buffer[1] = '0';
			uitoa_dec(buffer + 2, global_record.date.year, 2);
			buffer[4] = '-';
			uitoa_dec(buffer + 5, global_record.date.month, 2);
			buffer[7] = '-';
			uitoa_dec(buffer + 8, global_record.date.day, 2);
			buffer[10] = ' ';
			uitoa_dec(buffer + 11, global_record.date.hours, 2);
			buffer[13] = ':';
			uitoa_dec(buffer + 14, global_record.date.minutes, 2);
			buffer[16] = '\0';
			display_4bit_set_position(0, 0);
			display_4bit_write_string(buffer);

			// Draw the test type.
			switch (global_record.type)
			{
			case TEST_TYPE_JUMP_SINGLE:
				buffer[0] = 'J'; buffer[1] = 'S'; buffer[2] = '\0';
				display_4bit_set_position(0, 1);
				display_4bit_write_string(buffer);
				jump_long_to_string(buffer, 17, distances[0]);
				display_4bit_set_position(8, 1);
				display_4bit_write_string(buffer);
				display_4bit_write_string(" cm");
				break;

			case TEST_TYPE_JUMP_MULTIPLE:
				buffer[0] = 'J'; buffer[1] = 'R'; buffer[2] = '\0';
				display_4bit_set_position(0, 1);
				display_4bit_write_string(buffer);

				// Find the BEST result...
				best_result = 0;
				for (i = 0; i < global_record.num_times; ++i) {
					if (distances[i] > distances[best_result]) {
						best_result = i;
					}
				}
				
				// Write out the best index and total number of results.
				display_4bit_set_position(3, 1);
				uitoa_dec(buffer, best_result + 1, 2);
				buffer[2] = '\0';
				display_4bit_write_string(buffer);
				display_4bit_write_character('/');
				uitoa_dec(buffer, global_record.num_times, 2);
				buffer[2] = '\0';
				display_4bit_write_string(buffer);

				// Convert the result to a string for display.
				jump_long_to_string(buffer, 6, distances[best_result]);
				display_4bit_set_position(9, 1);
				display_4bit_write_string(buffer);
				display_4bit_write_string("cm");
				break;

			case TEST_TYPE_SPRINT_SENSOR:
				buffer[0] = 'S'; buffer[1] = 'S'; buffer[2] = '\0';
				display_4bit_set_position(0, 1);
				display_4bit_write_string(buffer);

				// Write out the last result (the total time).
				display_4bit_set_position(3, 1);
				buffer[0] = 'S'; buffer[1] = ':';
				uitoa_dec(buffer + 2, global_record.num_times, 2);
				buffer[4] = '\0';
				display_4bit_write_string(buffer);

				// Convert the result to a string for display.
				time_long_to_string(buffer, 17, results[global_record.num_times - 1]);
				display_4bit_set_position(8, 1);
				display_4bit_write_string(buffer);
				break;

			case TEST_TYPE_SPRINT_SIGNAL:
				buffer[0] = 'S'; buffer[1] = 'I'; buffer[2] = '\0';
				display_4bit_set_position(0, 1);
				display_4bit_write_string(buffer);
				
				// Write out the last result (the total time).
				display_4bit_set_position(3, 1);
				buffer[0] = 'S'; buffer[1] = ':';
				uitoa_dec(buffer + 2, global_record.num_times, 2);
				buffer[4] = '\0';
				display_4bit_write_string(buffer);

				// Convert the result to a string for display.
				time_long_to_string(buffer, 17, results[global_record.num_times - 1]);
				display_4bit_set_position(8, 1);
				display_4bit_write_string(buffer);
				break;

			case TEST_TYPE_SPRINT_RANDOM:
				buffer[0] = 'S'; buffer[1] = 'R'; buffer[2] = '\0';
				display_4bit_set_position(0, 1);
				display_4bit_write_string(buffer);
				
				// Write out the last result (the total time).
				display_4bit_set_position(3, 1);
				buffer[0] = 'S'; buffer[1] = ':';
				uitoa_dec(buffer + 2, global_record.num_times, 2);
				buffer[4] = '\0';
				display_4bit_write_string(buffer);

				// Convert the result to a string for display.
				time_long_to_string(buffer, 17, results[global_record.num_times - 1]);
				display_4bit_set_position(8, 1);
				display_4bit_write_string(buffer);
				break;

			case TEST_TYPE_SPRINT_SWITCH:
				buffer[0] = 'S'; buffer[1] = 'W'; buffer[2] = '\0';
				display_4bit_set_position(0, 1);
				display_4bit_write_string(buffer);
				
				// Write out the last result (the total time).
				display_4bit_set_position(3, 1);
				buffer[0] = 'S'; buffer[1] = ':';
				uitoa_dec(buffer + 2, global_record.num_times, 2);
				buffer[4] = '\0';
				display_4bit_write_string(buffer);

				// Convert the result to a string for display.
				time_long_to_string(buffer, 17, results[global_record.num_times - 1]);
				display_4bit_set_position(8, 1);
				display_4bit_write_string(buffer);
				break;
			
			case TEST_TYPE_ENDURANCE_MULTI:
				buffer[0] = 'M'; buffer[1] = 'S'; buffer[2] = '\0';
				display_4bit_set_position(0, 1);
				display_4bit_write_string(buffer);

				// Find the BEST result (lowest value in 'results' array)...
				best_result = 1;
				for (i = 1; i < global_record.num_times; ++i) {
					if (results[i] < results[best_result]) {
						best_result = i;
					}
				}

				// Convert the result to a string for display.
				time_long_to_string(buffer, 17, results[best_result]);
				display_4bit_set_position(8, 1);
				display_4bit_write_string(buffer);

				// Write out the best index and total number of results.
				// note: We overwritr the highest digit of the timing result!
				display_4bit_set_position(3, 1);
				uitoa_dec(buffer, best_result, 2);
				buffer[2] = '\0';
				display_4bit_write_string(buffer);

				display_4bit_write_character('/');
				uitoa_dec(buffer, global_record.num_times, 2);
				buffer[2] = ' ';
				buffer[3] = '\0';
				display_4bit_write_string(buffer);
				break;

			case TEST_TYPE_ENDURANCE_BEEP:
				buffer[0] = 'B'; buffer[1] = 'T'; buffer[2] = '\0';
				display_4bit_set_position(0, 1);
				display_4bit_write_string(buffer);

				display_4bit_set_position(3, 1);
				display_4bit_write_string("STG:");
				uitoa_dec(buffer, (unsigned int) results[0], 2);
				buffer[2] = '\0';
				display_4bit_write_string(buffer);
				
				display_4bit_write_string(" LVL:");
				uitoa_dec(buffer, (unsigned int) distances[0], 2);
				buffer[2] = '\0';
				display_4bit_write_string(buffer);
				break;
			};

			redraw = 0;
		}

		delay_ms(100);
	}

	return 0;
}

void time_long_to_string (unsigned char *buf, unsigned char buf_len, unsigned long time)
{
	unsigned int minutes, seconds, hundreds, temp;
	
	#ifdef DEBUG_LONG_TO_STRING
	unsigned char buffer[32];
	#endif

	#ifdef DEBUG_LONG_TO_STRING
	sprintf32(buffer, "time: %u\r\n", time);
	uart_send_string(buffer);
	#endif

	minutes = (unsigned int) (time / 60000);
	temp = time - (minutes * 60000);
	seconds = (unsigned int) (temp / 1000);
	temp = temp - (seconds * 1000);
	hundreds = (unsigned int) (temp / 10);

	#ifdef DEBUG_LONG_TO_STRING
	sprintf(buffer, "minutes: %u\r\n", minutes);
	uart_send_string(buffer);
	sprintf(buffer, "seconds: %u\r\n", seconds);
	uart_send_string(buffer);
	sprintf(buffer, "hundreds: %u\r\n", hundreds);
	uart_send_string(buffer);
	#endif

	uitoa_dec(buf, minutes, 2);
	*(buf + 2) = ':';

	uitoa_dec(buf + 3, seconds, 2);
	*(buf + 5) = ',';

	uitoa_dec(buf + 6, hundreds, 2);
	*(buf + 8) = '\0';
}


void jump_long_to_string (unsigned char *buf, unsigned char buf_len, unsigned long height_mm)
{
	unsigned int centimeters;

	centimeters = (unsigned int) (height_mm / 10);
	height_mm -= (centimeters * 10);
	
	uitoa_dec(buf, centimeters, 3);
	*(buf + 3) = ',';
	uitoa_dec(buf + 4, height_mm, 1);
	*(buf + 5) = '\0';
}


/*
Function: chache_results

Description: Writes a list of all record addresses into 'dst' whose
    'player_id' and 'type' matches the correspondingly named parameters
    to the function. The number of records written is returned, and
    the function will never attempt to write more than 'dst_len' entries.
*/
unsigned char cache_results (unsigned int *dst,
	test_types_t type, unsigned char player_id)
{
	unsigned int end, size;
	unsigned char i;

	// Read ot the value indicating the end of the memory.
	fm3164_read_memory(0x0000, (unsigned char *) &end, sizeof(unsigned int));

	current_record = 0x0002;
	i = 0;
	while (current_record < end && i < NUM_CACHED_RESULTS) {
		size = record_retrieve();
		if (global_record.player_id == player_id && (global_record.type & type)) {
			dst[i++] = current_record - size;
		}	
	}

	return i;
}

keys_t get_input_digit (unsigned int *result)
{
	keys_t key;
	unsigned char value, result_value;
	unsigned long timer;

	result_value = *result;

	timer = 0;
	wait_for_keypad_release();

	// The input loop is terminated by pressing 'ok', 'next' or timeout expiration.
	while (timer < 100000) {
		key = keypad_scan();

		value = 10; // This is an invalid number.

		switch (key)
		{
		case KEY_0:
			value = 0;
			break;

		case KEY_1:
			value = 1;
			break;

		case KEY_2:
			value = 2;
			break;

		case KEY_3:
			value = 3;
			break;

		case KEY_4:
			value = 4;
			break;

		case KEY_5:
			value = 5;
			break;

		case KEY_6:
			value = 6;
			break;

		case KEY_7:
			value = 7;
			break;

		case KEY_8:
			value = 8;
			break;

		case KEY_9:
			value = 9;
			break;
	
		default:
			return key;
			break;
		};

		// If value have been modified someone has input a digit. Process it.
		if (value != 10) {
			result_value *= 10;
			result_value += value;
			*result = result_value;
			return key;
		}

		++timer;
	}

	return KEY_NONE;
}

/*
const unsigned char *keycode_to_string (keys_t key)
{
	switch (key)
	{
	case KEY_NONE:
		return "KEY_NONE";
		break;
	case KEY_ON:
		break;

	case KEY_OFF:
		return "KEY_OFF";
		break;
	case KEY_CANCEL:
		return "KEY_CANCEL";
		break;
	case KEY_PREV:
		return "KEY_PREV";
		break;

	case KEY_NEXT:
		return "KEY_NEXT";
		break;
	case KEY_SUB_INC:
		return "KEY_SUB_INC";
		break;
	case KEY_POP_DEC:
		return "KEY_POP_DEC";
		break;
	case KEY_OK:
		return "KEY_OK";
		break;
	case KEY_PRINT:
		return "KEY_PRINT";
		break;
	case KEY_1:
		return "KEY_1";
		break;
	case KEY_2:
		return "KEY_2";
		break;
	case KEY_3:
		return "KEY_3";
		break;
	case KEY_4:
		return "KEY_4";
		break;
	case KEY_5:
		return "KEY_5";
		break;
	case KEY_6:
		return "KEY_6";
		break;
	case KEY_7:
		return "KEY_7";
		break;
	case KEY_8:
		return "KEY_8";
		break;
	case KEY_9:
		return "KEY_9";
		break;
	case KEY_0:
		return "KEY_10";
		break;
	};
}
*/

/*
Function: airtime_to_height
Description: Converts an time in milliseconds to a jump height based on
    Earths gravitational acceleration. The result is in millimetres.
*/
unsigned long airtime_to_height (unsigned long time_ms)
{

}

/*
Function: menu_record_remove
Description: Invoked from the result menu in order to delete a specifik result.
*/
unsigned char menu_record_remove (unsigned int address)
{
	keys_t key;

	display_4bit_clear();
	display_4bit_set_position(0, 0);
	display_4bit_write_string("DELETE?");

	wait_for_keypad_release();

	while (1) {
		key = keypad_scan();
		if (key == KEY_CANCEL) {
			return 0;
		} else if (key == KEY_OK) {
			break;
		}
	}

	record_remove(address);

	return 1;
}

/*
Function: record_remove
Description: Removes the the record at 'address' from memory. The 'end' pointer is
    automatically updated at EEPROM address 0x0000.
*/
void record_remove (unsigned int address)
{
	unsigned int size_of_record, dst, src, end;
	unsigned char temp;
	record_t local_record;

	// Read out the size of the memory contents.
	fm3164_read_memory(0x0000, (unsigned char *) &end, sizeof(end));

	fm3164_read_memory(address, (unsigned char *) &local_record, sizeof(local_record));

	//record_load(address, &local_record, &local_results[0], &local_distances[0]);

	size_of_record = sizeof(local_record) + local_record.num_times * 
		(sizeof(unsigned long) + sizeof(unsigned int));
	dst = address;
	src = address + size_of_record;

	// Overwrite the record with the rest of the memory.
	while (src < end) {
		fm3164_read_memory_byte(src++, &temp);
		fm3164_write_byte_memory(dst++, temp);
	}
	
	end -= size_of_record;
	fm3164_write_memory(0x0000, (unsigned char *) &end, sizeof(end));
}