#ifndef TIMEIT_MENU2
#define TIMEIT_MENU2

#include "keypad.h"
#include "display_4bit.h"
#include "ramtron.h"
#include "util.h"
#include <stdio.h>
#include <memory.h>

#pragma CLOCK_FREQ 40000000

#define TIMEIT_MAX_PLAYERS 100
#define MAX_SENSORS 64
#define MIN_SENSORS 1
#define PLAYER_ID_MIN 0
#define PLAYER_ID_MAX 99
#define PLAYER_COUNT 100
#define KEY_PRESS_DELAY 60
#define DISTANCE_MIN 1
#define DISTANCE_MAX 999
#define START_BEEP_DELAY 200
#define NOTICE_BEEP_DELAY 200
#define TIMESTAMP_BEEP_DELAY 200
#define NUM_CACHED_RESULTS 50

#define ENABLE_BEEP_RECORDS 1 

// Debug definitions (uncomment to enable UART debug info)
//#define DEBUG 1
//#define DEBUG_RESULTS 1
//#define MENU_DEBUG 1
//#define DEBUG_WRITE_RECORDS 1
//#define DEBUG_JUMP_SINGLE 1
//#define DEBUG_JUMP_MULTIPLE 1
//#define DEBUG_LONG_TO_STRING 1
//#define DEBUG_EEPROM 1
//#define DEBUG_INPUT_DIGIT 1
//#define DEBUG_ENDURANCE_MULTI 1
//#define DEBUG_ENDURANCE_BEEP 1
//#define DEBUG_ENDURANCE_SETUP 1

/*
typedef enum system_states
{
	SYSTEM_STATE_IDLE,
	SYSTEM_STATE_JUMP_SINGLE_STEP1,
	SYSTEM_STATE_JUMP_SINGLE_STEP2,
	SYSTEM_STATE_JUMP_SINGLE_STEP3,
} system_states_t;
*/

typedef enum test_types
{
	TEST_TYPE_JUMP_SINGLE = 0x01,
	TEST_TYPE_JUMP_MULTIPLE = 0x02,
	TEST_TYPE_SPRINT_SENSOR = 0x04,
	TEST_TYPE_SPRINT_SIGNAL = 0x08,
	TEST_TYPE_SPRINT_RANDOM = 0x10,
	TEST_TYPE_SPRINT_SWITCH = 0x20,
	TEST_TYPE_ENDURANCE_BEEP = 0x40,
	TEST_TYPE_ENDURANCE_MULTI = 0x80
} test_types_t;


typedef enum menu_states
{
	MENU_STATE_SPRINT,
	MENU_STATE_SPRINT_SENSOR,
	MENU_STATE_SPRINT_SIGNAL,
	MENU_STATE_SPRINT_RANDOM,
	MENU_STATE_SPRINT_SWITCH,
	MENU_STATE_JUMP,
	MENU_STATE_JUMP_SINGLE,
	MENU_STATE_JUMP_MULTIPLE,
	MENU_STATE_ENDURANCE,
	MENU_STATE_ENDURANCE_BEEP,
	MENU_STATE_ENDURANCE_MULTI,
	MENU_STATE_RESULTS,
	MENU_STATE_RESULTS_SPRINT,
	MENU_STATE_RESULTS_JUMP,
	MENU_STATE_RESULTS_ENDURANCE,
	MENU_STATE_SETUP,
	MENU_STATE_SETUP_DATE,
	MENU_STATE_SETUP_MEMORY,
	MENU_STATE_SETUP_MEMORY_STATUS,
	MENU_STATE_SETUP_MEMORY_CLEARPLAYER,
	MENU_STATE_SETUP_MEMORY_CLEAR,
	MENU_STATE_SETUP_SOFTWARE,
	MENU_STATE_SETUP_SOFTWARE_DISPLAY,
	MENU_STATE_DOWNLOAD_TRANSMITTING,
} menu_states_t;


/*
structure: record
    player_id: 8-bit unsigned integer indicating the players identity.
	test_id: 16-bit unsigned integer indicating the test's identity.
    date: fm3164_t structure indicating the time for the test.
    type: enumeration (assumed 8-bit unsigned int) specifying the test type.
    num_times: 8-bit unsigned integer specifying the number of times.
*/
typedef struct record
{
	unsigned char player_id;
	//unsigned int test_id;
	fm3164_date_t date;
	test_types_t type;
	unsigned char num_times;
} record_t;

unsigned char record_store (void);
unsigned int record_retrieve (void);
void record_retrieve_previous (void);


void menu_process (keys_t key);
void menu_draw (void);
void date_modify (void);
void init_eeprom (void);
void transmit_records (void);

void realtime_update (unsigned long result, unsigned char pos_x, unsigned char pos_y);

// Functions imlementing the logic for a test.
unsigned char test_jump_single (void);
unsigned char test_jump_multiple (void);
unsigned char test_endurance_beep (void);
unsigned char test_endurance_multi (void);
unsigned char test_sprint_sensor (void);
unsigned char test_sprint_signal (void);
unsigned char test_sprint_random (void);

// Setup functions (contains logic allowing user to input a test config)
unsigned char setup_sprint_test (unsigned char reuse, test_types_t test_type);
unsigned char setup_endurance_test (unsigned char reuse);
unsigned char setup_jump_test (unsigned char reuse);

void wait_for_keypad_release (void);
unsigned char prompt_for_save (void);
void record_remove_player (unsigned char player_id);
void remove_result (unsigned int record, unsigned char result);
void record_remove (unsigned int address);
unsigned char menu_record_remove (unsigned int address);

void menu_setup_date (void);
void menu_clear_memory (void);
unsigned char menu_clear_player (void);
void menu_memory_status (void);

void menu_results_sprint (void);
void menu_results_jump (void);
void menu_results_endurance (void);

//keys_t input_player (unsigned char *player_id);
unsigned char display_time_results (test_types_t type, unsigned char player_id);

unsigned long airtime_to_height (unsigned long time_ms);

#ifdef DEBUG
void dump_memory_to_serial (test_types_t type);
#endif

void time_long_to_string (unsigned char *buf, unsigned char buf_len, unsigned long time);
void jump_long_to_string (unsigned char *buf, unsigned char buf_len, unsigned long height_mm);
unsigned char cache_results (unsigned int *dst, test_types_t type, unsigned char player_id);
/*unsigned char input_numeric (unsigned char *result, unsigned char digits,
	unsigned int min, unsigned int max);*/
//keys_t get_input_digit (unsigned int *result);
//const unsigned char *keycode_to_string (keys_t key);

#endif
