#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

const char *data =
	"PLAYERID:1;TESTID:1;TYPE:SW;DATE:20080501133512;"
	"RESULTS:021221,024121,041331;"
	"PLAYERID:1;TESTID:1;TYPE:SW;DATE:20080501133512;"
	"RESULTS:021221,024121,041331;";
	

const char *config_string = "baud=9600 parity=N data=8 stop=1 to=on xon=off dtr=off rts=off";

int main (int argc, char *argv[])
{
	HANDLE serial_port;
	DCB original_config, config;
	COMMTIMEOUTS original_timeouts, timeouts;

	char buffer[32];
	int buffer_length;
	int bytes_written;

	if (argc < 2) {
		fprintf(stderr, "usage: timeitsim <PORT>\n");
		exit(2);
	}

	memset(&config, 0, sizeof(DCB));
	memset(&original_config, 0, sizeof(DCB));
	memset(&timeouts, 0, sizeof(COMMTIMEOUTS));
	memset(&original_timeouts, 0, sizeof(COMMTIMEOUTS));

	// Open the serial port for writing. The name of the port is supposedly
	// the first argument to the program.
	if ((serial_port = CreateFile(argv[1], GENERIC_WRITE | GENERIC_READ,
		0, 0, OPEN_EXISTING, 0, 0)) == 0) {
		fprintf(stderr, "error: Unable to open serial port \"%s\".\n",
			argv[1]);
	}

	printf("Retrieving existing serial port configuration...");
	if (!GetCommState(serial_port, &original_config)) {
		fprintf(stderr, "error: Unable to retrieve port "
			"communications configuration.\n");
		CloseHandle(serial_port);
		exit(2);
	}

	if (!GetCommTimeouts(serial_port, &original_timeouts)) {
		fprintf(stderr, "error: Unable to retrieve port "
			"timeouts configuration.\n");
		CloseHandle(serial_port);
		exit(2);
	}

	printf(" done\nCreating new configuration...");

	if (!BuildCommDCBAndTimeouts(config_string, &config, &timeouts)) {
		fprintf(stderr, "error: Unable to create new port configuration.\n");
		CloseHandle(serial_port);
		exit(2);
	}

	printf(" done\nApplying new configuration...");

	if (!SetCommState(serial_port, &config)) {
		fprintf(stderr, "error: Unable to set new port "
			"communications configuraton.\n");
		SetCommState(serial_port, &original_config);
		CloseHandle(serial_port);
		exit(2);
	}

	if (!SetCommTimeouts(serial_port, &timeouts)) {
		fprintf(stderr, "error: Unable to set new port "
			"timeouts configuraton.\n");
		CloseHandle(serial_port);
		exit(2);
	}

	/*
	printf(" done\nWriting data to serial port...");
	buffer_length = strlen(data);
	bytes_written = 0;
	if (!WriteFile(serial_port, data, buffer_length, &bytes_written, 0)) {
		SetCommState(serial_port, &original_config);
		SetCommTimeouts(serial_port, &original_timeouts);
		CloseHandle(serial_port);
		fprintf(stderr, "error: Unable to write data to serial port.");
		fprintf(stderr, "Error code: %d\n", GetLastError());
		exit(2);
	}
	*/

	buffer_length = sizeof(buffer);
	bytes_written = 0;
	if (ReadFile(serial_port, buffer, buffer_length, &bytes_written, 0) == 0) {
		fprintf(stderr, "error: Unable to read data from port. "
			"Error code is %d", GetLastError());
		exit(2);
	}

	printf("Buffer contents: \"%s\"\n", buffer);

	printf(" done\nBytes written: %d\nRestoring original "
		"port configuration...", bytes_written);

	if (!SetCommState(serial_port, &original_config)) {
		fprintf(stderr, "error: Unable to restore original port "
			"communication configuration.\n");
		CloseHandle(serial_port);
		exit(2);
	}

	if (!SetCommTimeouts(serial_port, &original_timeouts)) {
		fprintf(stderr, "error: Unable to restore original port "
			"timeouts configuration.\n");
		CloseHandle(serial_port);
		exit(2);
	}

	printf(" done\nClosing port...");

	if (!CloseHandle(serial_port)) {
		fprintf(stderr, "error: Unable to close serial port.\n");
		exit(2);
	}

	printf("done\n\nAll done!\n");

	return 0;
}


